# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Classes to handle AEA configurations."""
import pprint
from abc import ABC
from collections import OrderedDict
from copy import copy, deepcopy
from operator import attrgetter
from pathlib import Path
from typing import (
    Any,
    Collection,
    Dict,
    FrozenSet,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import packaging
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from aea.__version__ import __version__ as __aea_version__
from aea.configurations.constants import (
    CONNECTIONS,
    CONTRACTS,
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_CONNECTION_CONFIG_FILE,
    DEFAULT_CONTRACT_CONFIG_FILE,
    DEFAULT_FINGERPRINT_IGNORE_PATTERNS,
    DEFAULT_LICENSE,
    DEFAULT_LOGGING_CONFIG,
    DEFAULT_PROTOCOL_CONFIG_FILE,
    DEFAULT_REGISTRY_NAME,
    DEFAULT_SKILL_CONFIG_FILE,
    DEFAULT_VERSION,
    PACKAGE_PUBLIC_ID_VAR_NAME,
    PROTOCOLS,
    SKILLS,
)
from aea.configurations.data_types import (
    CRUDCollection,
    ComponentId,
    ComponentType,
    Dependencies,
    Dependency,
    JSONSerializable,
    PackageId,
    PackageType,
    PackageVersion,
    PublicId,
)
from aea.configurations.validation import ConfigValidator, validate_data_with_pattern
from aea.exceptions import enforce
from aea.helpers.base import (
    CertRequest,
    SimpleId,
    SimpleIdOrStr,
    load_module,
    recursive_update,
)
from aea.helpers.ipfs.base import IPFSHashOnly


# for tests
_ = [PackageId, PackageVersion]


T = TypeVar("T")


def dependencies_from_json(obj: Dict[str, Dict]) -> Dependencies:
    """
    Parse a JSON object to get an instance of Dependencies.

    :param obj: a dictionary whose keys are package names and values are dictionary with package specifications.
    :return: a Dependencies object.
    """
    return {key: Dependency.from_json({key: value}) for key, value in obj.items()}


def dependencies_to_json(dependencies: Dependencies) -> Dict[str, Dict]:
    """
    Transform a Dependencies object into a JSON object.

    :param dependencies: an instance of "Dependencies" type.
    :return: a dictionary whose keys are package names and
             values are the JSON version of a Dependency object.
    """
    result = {}
    for key, value in dependencies.items():
        dep_to_json = value.to_json()
        package_name = list(dep_to_json.items())[0][0]
        enforce(
            key == package_name, f"Names of dependency differ: {key} != {package_name}"
        )
        result[key] = dep_to_json[key]
    return result


def _get_default_configuration_file_name_from_type(
    item_type: Union[str, PackageType]
) -> str:
    """Get the default configuration file name from item type."""
    item_type = PackageType(item_type)
    if item_type == PackageType.AGENT:
        return DEFAULT_AEA_CONFIG_FILE
    if item_type == PackageType.PROTOCOL:
        return DEFAULT_PROTOCOL_CONFIG_FILE
    if item_type == PackageType.CONNECTION:
        return DEFAULT_CONNECTION_CONFIG_FILE
    if item_type == PackageType.SKILL:
        return DEFAULT_SKILL_CONFIG_FILE
    if item_type == PackageType.CONTRACT:
        return DEFAULT_CONTRACT_CONFIG_FILE
    raise ValueError(  # pragma: no cover
        "Item type not valid: {}".format(str(item_type))
    )


class ProtocolSpecificationParseError(Exception):
    """Exception for parsing a protocol specification file."""


class Configuration(JSONSerializable, ABC):
    """Configuration class."""

    __slots__ = ("_key_order",)

    def __init__(self) -> None:
        """Initialize a configuration object."""
        # a list of keys that remembers the key order of the configuration file.
        # this is set by the configuration loader.
        self._key_order: List[str] = []

    @classmethod
    def from_json(cls, obj: Dict) -> "Configuration":
        """Build from a JSON object."""

    @property
    def ordered_json(self) -> OrderedDict:
        """
        Reorder the dictionary according to a key ordering.

        This method takes all the keys in the key_order list and
        get the associated value in the dictionary (if present).
        For the remaining keys not considered in the order,
        it will use alphanumerical ordering.

        In particular, if key_order is an empty sequence, this reduces to
        alphanumerical sorting.

        It does not do side-effect.
        :return: the ordered dictionary.
        """
        data = self.json
        result = OrderedDict()  # type: OrderedDict

        # parse all the known keys. This might ignore some keys in the dictionary.
        seen_keys = set()
        for key in self._key_order:
            enforce(key not in result, "Key in results!")
            value = data.get(key)
            if value is not None:
                result[key] = value
                seen_keys.add(key)

        # Now process the keys in the dictionary that were not covered before.
        for key, value in data.items():
            if key not in seen_keys:
                result[key] = value
        return result


class PackageConfiguration(Configuration, ABC):
    """
    This class represent a package configuration.

    A package can be one of:
    - agents
    - protocols
    - connections
    - skills
    - contracts
    """

    __slots__ = (
        "_name",
        "_author",
        "version",
        "license",
        "fingerprint",
        "fingerprint_ignore_patterns",
        "build_entrypoint",
        "_aea_version",
        "_aea_version_specifiers",
        "_directory",
    )

    default_configuration_filename: str
    package_type: PackageType
    FIELDS_ALLOWED_TO_UPDATE: FrozenSet[str] = frozenset(["build_directory"])
    schema: str
    CHECK_EXCLUDES: List[Tuple[str]] = []

    def __init__(
        self,
        name: SimpleIdOrStr,
        author: SimpleIdOrStr,
        version: str = "",
        license_: str = "",
        aea_version: str = "",
        fingerprint: Optional[Dict[str, str]] = None,
        fingerprint_ignore_patterns: Optional[Sequence[str]] = None,
        build_entrypoint: Optional[str] = None,
    ) -> None:
        """
        Initialize a package configuration.

        :param name: the name of the package.
        :param author: the author of the package.
        :param version: the version of the package (SemVer format).
        :param license_: the license.
        :param aea_version: either a fixed version, or a set of specifiers
           describing the AEA versions allowed.
           (default: empty string - no constraint).
           The fixed version is interpreted with the specifier '=='.
        :param fingerprint: the fingerprint.
        :param fingerprint_ignore_patterns: a list of file patterns to ignore files to fingerprint.
        :param build_entrypoint: path to a script to execute at build time.
        """
        super().__init__()
        if name is None or author is None:  # pragma: nocover
            raise ValueError("Name and author must be set on the configuration!")
        self._name = SimpleId(name)
        self._author = SimpleId(author)
        self.version = version if version != "" else DEFAULT_VERSION
        self.license = license_ if license_ != "" else DEFAULT_LICENSE
        self.fingerprint = fingerprint if fingerprint is not None else {}
        self.fingerprint_ignore_patterns = (
            fingerprint_ignore_patterns
            if fingerprint_ignore_patterns is not None
            else []
        )
        self.build_entrypoint = build_entrypoint
        self._aea_version = aea_version if aea_version != "" else __aea_version__
        self._aea_version_specifiers = self._parse_aea_version_specifier(aea_version)

        self._directory = None  # type: Optional[Path]

    @property
    def name(self) -> str:
        """Get the name."""
        return str(self._name)

    @name.setter
    def name(self, value: SimpleIdOrStr) -> None:
        """Set the name."""
        self._name = SimpleId(value)

    @property
    def author(self) -> str:
        """Get the author."""
        return str(self._author)

    @author.setter
    def author(self, value: SimpleIdOrStr) -> None:
        """Set the author."""
        self._author = SimpleId(value)

    @property
    def aea_version(self) -> str:
        """Get the 'aea_version' attribute."""
        return self._aea_version

    @aea_version.setter
    def aea_version(self, new_aea_version: str) -> None:
        """Set the 'aea_version' attribute."""
        self._aea_version_specifiers = self._parse_aea_version_specifier(
            new_aea_version
        )
        self._aea_version = new_aea_version

    def check_aea_version(self) -> None:
        """
        Check that the AEA version matches the specifier set.

        :raises ValueError if the version of the aea framework falls within a specifier.
        """
        _check_aea_version(self)

    @property
    def directory(self) -> Optional[Path]:
        """Get the path to the configuration file associated to this file, if any."""
        return self._directory

    @directory.setter
    def directory(self, directory: Path) -> None:
        """Set directory if not already set."""
        if self._directory is not None:  # pragma: nocover
            raise ValueError("Directory already set")
        self._directory = directory

    @property
    def package_id(self) -> PackageId:
        """Get package id."""
        return PackageId(package_type=self.package_type, public_id=self.public_id)

    @staticmethod
    def _parse_aea_version_specifier(aea_version_specifiers: str) -> SpecifierSet:
        try:
            Version(aea_version_specifiers)
            return SpecifierSet("==" + aea_version_specifiers)
        except packaging.version.InvalidVersion:
            pass
        return SpecifierSet(aea_version_specifiers)

    @property
    def aea_version_specifiers(self) -> SpecifierSet:
        """Get the AEA version set specifier."""
        return self._aea_version_specifiers

    @property
    def public_id(self) -> PublicId:
        """Get the public id."""
        return PublicId(self.author, self.name, self.version)

    @property
    def package_dependencies(self) -> Set[ComponentId]:
        """Get the package dependencies."""
        return set()

    def update(self, data: Dict, env_vars_friendly: bool = False) -> None:
        """
        Update configuration with other data.

        :param data: the data to replace.
        :return: None
        """
        if not data:  # do nothing if nothing to update
            return

        self.check_overrides_valid(data, env_vars_friendly=env_vars_friendly)
        self._create_or_update_from_json(
            obj=self.make_resulting_config_data(data), instance=self
        )

    @classmethod
    def validate_config_data(
        cls, json_data: Dict, env_vars_friendly: bool = False
    ) -> None:
        """Perform config validation."""
        ConfigValidator(cls.schema, env_vars_friendly=env_vars_friendly).validate(
            json_data
        )

    @classmethod
    def from_json(cls, obj: Dict) -> "PackageConfiguration":
        """Initialize from a JSON object."""
        return cls._create_or_update_from_json(obj=obj, instance=None)

    @classmethod
    def _create_or_update_from_json(
        cls, obj: Dict, instance: Any = None
    ) -> "PackageConfiguration":
        """Create new config object or updates existing one from json data."""
        raise NotImplementedError  # pragma: nocover

    def make_resulting_config_data(self, overrides: Dict) -> Dict:
        """Make config data with overrides applied.

        Does not update config, just creates json representation
        """
        current_config = self.json
        recursive_update(current_config, overrides, allow_new_values=True)
        return current_config

    def check_overrides_valid(
        self, overrides: Dict, env_vars_friendly: bool = False
    ) -> None:
        """Check overrides is correct, return list of errors if present."""
        # check for permited overrides
        self._check_overrides_corresponds_to_overridable(
            overrides, env_vars_friendly=env_vars_friendly
        )
        # check resulting config with applied overrides passes validation

        result_config = self.make_resulting_config_data(overrides)
        self.validate_config_data(result_config, env_vars_friendly=env_vars_friendly)

    def _check_overrides_corresponds_to_overridable(
        self, overrides: Dict, env_vars_friendly: bool = False
    ) -> None:
        """Check overrides is correct, return list of errors if present."""
        errors_list = validate_data_with_pattern(
            overrides,
            self.get_overridable(),
            excludes=self.CHECK_EXCLUDES,
            skip_env_vars=env_vars_friendly,
        )
        if errors_list:
            raise ValueError(errors_list[0])

    def get_overridable(self) -> dict:
        """Get dictionary of values that can be updated for this config."""
        return {k: self.json.get(k) for k in self.FIELDS_ALLOWED_TO_UPDATE}

    @classmethod
    def _apply_params_to_instance(
        cls, params: dict, instance: Optional["PackageConfiguration"]
    ) -> "PackageConfiguration":
        """Constructs or update instance with params provided."""
        directory = (
            instance.directory if instance and hasattr(instance, "directory") else None
        )

        if instance is None:
            instance = cls(**params)
        else:
            instance.__init__(**params)  # type: ignore

        if directory and not instance.directory:
            instance.directory = directory

        return instance


class ComponentConfiguration(PackageConfiguration, ABC):
    """Class to represent an agent component configuration."""

    package_type: PackageType

    __slots__ = ("pypi_dependencies", "_build_directory")

    def __init__(
        self,
        name: SimpleIdOrStr,
        author: SimpleIdOrStr,
        version: str = "",
        license_: str = "",
        aea_version: str = "",
        fingerprint: Optional[Dict[str, str]] = None,
        fingerprint_ignore_patterns: Optional[Sequence[str]] = None,
        build_entrypoint: Optional[str] = None,
        build_directory: Optional[str] = None,
        dependencies: Optional[Dependencies] = None,
    ) -> None:
        """Set component configuration."""
        super().__init__(
            name,
            author,
            version,
            license_,
            aea_version,
            fingerprint,
            fingerprint_ignore_patterns,
            build_entrypoint,
        )
        self.pypi_dependencies: Dependencies = dependencies if dependencies is not None else {}
        self._build_directory = build_directory

    @property
    def build_directory(self) -> Optional[str]:
        """Get the component type."""
        return self._build_directory

    @build_directory.setter
    def build_directory(self, value: Optional[str]) -> None:
        """Get the component type."""
        self._build_directory = value

    @property
    def component_type(self) -> ComponentType:
        """Get the component type."""
        return ComponentType(self.package_type.value)

    @property
    def component_id(self) -> ComponentId:
        """Get the component id."""
        return ComponentId(self.component_type, self.public_id)

    @property
    def prefix_import_path(self) -> str:
        """Get the prefix import path for this component."""
        return "packages.{}.{}.{}".format(
            self.public_id.author, self.component_type.to_plural(), self.public_id.name
        )

    @property
    def is_abstract_component(self) -> bool:
        """Check whether the component is abstract."""
        return False

    def _check_configuration_consistency(self, directory: Path) -> None:
        """Check that the configuration file is consistent against a directory."""
        self.check_fingerprint(directory)
        self.check_aea_version()
        self.check_public_id_consistency(directory)

    def check_fingerprint(self, directory: Path) -> None:
        """
        Check that the fingerprint are correct against a directory path.

        :raises ValueError if:
            - the argument is not a valid package directory
            - the fingerprints do not match.
        """
        if not directory.exists() or not directory.is_dir():
            raise ValueError("Directory {} is not valid.".format(directory))
        _compare_fingerprints(
            self, directory, False, self.component_type.to_package_type()
        )

    def check_public_id_consistency(self, directory: Path) -> None:
        """
        Check that the public ids in the init file match the config.

        :raises ValueError if:
            - the argument is not a valid package directory
            - the public ids do not match.
        """
        if not directory.exists() or not directory.is_dir():
            raise ValueError("Directory {} is not valid.".format(directory))
        _compare_public_ids(self, directory)


class ConnectionConfig(ComponentConfiguration):
    """Handle connection configuration."""

    default_configuration_filename = DEFAULT_CONNECTION_CONFIG_FILE
    package_type = PackageType.CONNECTION
    schema = "connection-config_schema.json"

    FIELDS_ALLOWED_TO_UPDATE: FrozenSet[str] = frozenset(
        ["config", "cert_requests", "is_abstract", "build_directory"]
    )

    __slots__ = (
        "class_name",
        "protocols",
        "connections",
        "restricted_to_protocols",
        "excluded_protocols",
        "dependencies",
        "description",
        "config",
        "is_abstract",
        "cert_requests",
    )

    def __init__(
        self,
        name: SimpleIdOrStr = "",
        author: SimpleIdOrStr = "",
        version: str = "",
        license_: str = "",
        aea_version: str = "",
        fingerprint: Optional[Dict[str, str]] = None,
        fingerprint_ignore_patterns: Optional[Sequence[str]] = None,
        build_entrypoint: Optional[str] = None,
        build_directory: Optional[str] = None,
        class_name: str = "",
        protocols: Optional[Set[PublicId]] = None,
        connections: Optional[Set[PublicId]] = None,
        restricted_to_protocols: Optional[Set[PublicId]] = None,
        excluded_protocols: Optional[Set[PublicId]] = None,
        dependencies: Optional[Dependencies] = None,
        description: str = "",
        connection_id: Optional[PublicId] = None,
        is_abstract: bool = False,
        cert_requests: Optional[List[CertRequest]] = None,
        **config: Any,
    ) -> None:
        """Initialize a connection configuration object."""
        if connection_id is None:
            enforce(name != "", "Name or connection_id must be set.")
            enforce(author != "", "Author or connection_id must be set.")
            enforce(version != "", "Version or connection_id must be set.")
        else:
            enforce(
                name in ("", connection_id.name,),
                "Non matching name in ConnectionConfig name and public id.",
            )
            name = connection_id.name
            enforce(
                author in ("", connection_id.author,),
                "Non matching author in ConnectionConfig author and public id.",
            )
            author = connection_id.author
            enforce(
                version in ("", connection_id.version,),
                "Non matching version in ConnectionConfig version and public id.",
            )
            version = connection_id.version
        super().__init__(
            name,
            author,
            version,
            license_,
            aea_version,
            fingerprint,
            fingerprint_ignore_patterns,
            build_entrypoint,
            build_directory,
            dependencies,
        )
        self.class_name = class_name
        self.protocols = protocols if protocols is not None else set()
        self.connections = connections if connections is not None else set()
        self.restricted_to_protocols = (
            restricted_to_protocols if restricted_to_protocols is not None else set()
        )
        self.excluded_protocols = (
            excluded_protocols if excluded_protocols is not None else set()
        )
        self.dependencies = dependencies if dependencies is not None else {}
        self.description = description
        self.config = config if len(config) > 0 else {}
        self.is_abstract = is_abstract
        self.cert_requests = cert_requests

    @property
    def package_dependencies(self) -> Set[ComponentId]:
        """Get the connection dependencies."""
        return {
            ComponentId(ComponentType.PROTOCOL, protocol_id)
            for protocol_id in self.protocols
        }.union(
            {
                ComponentId(ComponentType.CONNECTION, connection_id)
                for connection_id in self.connections
            }
        )

    @property
    def is_abstract_component(self) -> bool:
        """Check whether the component is abstract."""
        return self.is_abstract

    @property
    def json(self) -> Dict:
        """Return the JSON representation."""
        result = OrderedDict(
            {
                "name": self.name,
                "author": self.author,
                "version": self.version,
                "type": self.component_type.value,
                "description": self.description,
                "license": self.license,
                "aea_version": self.aea_version,
                "fingerprint": self.fingerprint,
                "fingerprint_ignore_patterns": self.fingerprint_ignore_patterns,
                PROTOCOLS: sorted(map(str, self.protocols)),
                CONNECTIONS: sorted(map(str, self.connections)),
                "class_name": self.class_name,
                "config": self.config,
                "excluded_protocols": sorted(map(str, self.excluded_protocols)),
                "restricted_to_protocols": sorted(
                    map(str, self.restricted_to_protocols)
                ),
                "dependencies": dependencies_to_json(self.dependencies),
                "is_abstract": self.is_abstract,
            }
        )

        if self.cert_requests is not None:
            result["cert_requests"] = list(map(attrgetter("json"), self.cert_requests))
        if self.build_entrypoint:
            result["build_entrypoint"] = self.build_entrypoint
        if self.build_directory:
            result["build_directory"] = self.build_directory
        return result

    @classmethod
    def _create_or_update_from_json(
        cls, obj: Dict, instance: Optional["ConnectionConfig"] = None
    ) -> "ConnectionConfig":
        """Create new config object or updates existing one from json data."""
        obj = {**(instance.json if instance else {}), **copy(obj)}
        restricted_to_protocols = obj.get("restricted_to_protocols", set())
        restricted_to_protocols = {
            PublicId.from_str(id_) for id_ in restricted_to_protocols
        }
        excluded_protocols = obj.get("excluded_protocols", set())
        excluded_protocols = {PublicId.from_str(id_) for id_ in excluded_protocols}
        dependencies = dependencies_from_json(obj.get("dependencies", {}))
        protocols = {PublicId.from_str(id_) for id_ in obj.get(PROTOCOLS, set())}
        connections = {PublicId.from_str(id_) for id_ in obj.get(CONNECTIONS, set())}
        cert_requests = (
            [
                # notice: yaml.load resolves datetimes strings to datetime.datetime objects
                CertRequest.from_json(cert_request_json)
                for cert_request_json in obj["cert_requests"]
            ]
            if "cert_requests" in obj
            else None
        )

        params = dict(
            name=cast(str, obj.get("name")),
            author=cast(str, obj.get("author")),
            version=cast(str, obj.get("version")),
            license_=cast(str, obj.get("license")),
            aea_version=cast(str, obj.get("aea_version", "")),
            fingerprint=cast(Dict[str, str], obj.get("fingerprint")),
            fingerprint_ignore_patterns=cast(
                Sequence[str], obj.get("fingerprint_ignore_patterns")
            ),
            build_entrypoint=cast(Optional[str], obj.get("build_entrypoint")),
            build_directory=cast(Optional[str], obj.get("build_directory")),
            class_name=cast(str, obj.get("class_name")),
            protocols=cast(Set[PublicId], protocols),
            connections=cast(Set[PublicId], connections),
            restricted_to_protocols=cast(Set[PublicId], restricted_to_protocols),
            excluded_protocols=cast(Set[PublicId], excluded_protocols),
            dependencies=cast(Dependencies, dependencies),
            description=cast(str, obj.get("description", "")),
            is_abstract=obj.get("is_abstract", False),
            cert_requests=cert_requests,
            **cast(dict, obj.get("config", {})),
        )

        instance = cast(
            ConnectionConfig, cls._apply_params_to_instance(params, instance)
        )

        return instance


class ProtocolConfig(ComponentConfiguration):
    """Handle protocol configuration."""

    default_configuration_filename = DEFAULT_PROTOCOL_CONFIG_FILE
    package_type = PackageType.PROTOCOL
    schema = "protocol-config_schema.json"
    FIELDS_ALLOWED_TO_UPDATE: FrozenSet[str] = frozenset()

    __slots__ = ("dependencies", "description", "protocol_specification_id")

    def __init__(
        self,
        name: SimpleIdOrStr,
        author: SimpleIdOrStr,
        version: str = "",
        license_: str = "",
        fingerprint: Optional[Dict[str, str]] = None,
        fingerprint_ignore_patterns: Optional[Sequence[str]] = None,
        build_entrypoint: Optional[str] = None,
        build_directory: Optional[str] = None,
        aea_version: str = "",
        dependencies: Optional[Dependencies] = None,
        description: str = "",
        protocol_specification_id: Optional[str] = None,
    ) -> None:
        """Initialize a connection configuration object."""
        super().__init__(
            name,
            author,
            version,
            license_,
            aea_version,
            fingerprint,
            fingerprint_ignore_patterns,
            build_entrypoint,
            build_directory,
            dependencies,
        )
        self.dependencies = dependencies if dependencies is not None else {}
        self.description = description

        # temporary solution till all protocols updated
        if protocol_specification_id is not None:
            self.protocol_specification_id = PublicId.from_str(
                str(protocol_specification_id)
            )
        else:
            # make protocol specification same as protocol id
            self.protocol_specification_id = self.public_id

    @property
    def json(self) -> Dict:
        """Return the JSON representation."""
        result = OrderedDict(
            {
                "name": self.name,
                "author": self.author,
                "version": self.version,
                "protocol_specification_id": str(self.protocol_specification_id),
                "type": self.component_type.value,
                "description": self.description,
                "license": self.license,
                "aea_version": self.aea_version,
                "fingerprint": self.fingerprint,
                "fingerprint_ignore_patterns": self.fingerprint_ignore_patterns,
                "dependencies": dependencies_to_json(self.dependencies),
            }
        )
        if self.build_entrypoint:
            result["build_entrypoint"] = self.build_entrypoint
        if self.build_directory:
            result["build_directory"] = self.build_directory
        return result

    @classmethod
    def _create_or_update_from_json(
        cls, obj: Dict, instance: Optional["ProtocolConfig"] = None
    ) -> "ProtocolConfig":
        """Initialize from a JSON object."""
        obj = {**(instance.json if instance else {}), **copy(obj)}
        dependencies = dependencies_from_json(obj.get("dependencies", {}))
        params = dict(
            name=cast(str, obj.get("name")),
            author=cast(str, obj.get("author")),
            protocol_specification_id=cast(str, obj.get("protocol_specification_id")),
            version=cast(str, obj.get("version")),
            license_=cast(str, obj.get("license")),
            aea_version=cast(str, obj.get("aea_version", "")),
            fingerprint=cast(Dict[str, str], obj.get("fingerprint")),
            fingerprint_ignore_patterns=cast(
                Sequence[str], obj.get("fingerprint_ignore_patterns")
            ),
            build_entrypoint=cast(Optional[str], obj.get("build_entrypoint")),
            build_directory=cast(Optional[str], obj.get("build_directory")),
            dependencies=dependencies,
            description=cast(str, obj.get("description", "")),
        )
        instance = cast(ProtocolConfig, cls._apply_params_to_instance(params, instance))

        return instance


class SkillComponentConfiguration:
    """This class represent a skill component configuration."""

    __slots__ = ("class_name", "file_path", "args")

    def __init__(
        self, class_name: str, file_path: Optional[str] = None, **args: Any
    ) -> None:
        """
        Initialize a skill component configuration.

        :param skill_component_type: the skill component type.
        :param class_name: the class name of the component.
        :param args: keyword arguments.
        """
        self.class_name = class_name
        self.file_path: Optional[Path] = Path(file_path) if file_path else None
        self.args = args

    @property
    def json(self) -> Dict:
        """Return the JSON representation."""
        result = {"class_name": self.class_name, "args": self.args}
        if self.file_path is not None:
            result["file_path"] = str(self.file_path.as_posix())
        return result

    @classmethod
    def from_json(cls, obj: Dict) -> "SkillComponentConfiguration":
        """Initialize from a JSON object."""
        return cls._create_or_update_from_json(obj, instance=None)

    @classmethod
    def _create_or_update_from_json(
        cls, obj: Dict, instance: Optional["SkillComponentConfiguration"] = None
    ) -> "SkillComponentConfiguration":
        """Initialize from a JSON object."""
        obj = {**(instance.json if instance else {}), **copy(obj)}
        class_name = cast(str, obj.get("class_name"))
        file_path = cast(Optional[str], obj.get("file_path"))
        params = dict(class_name=class_name, file_path=file_path, **obj.get("args", {}))

        instance = cast(
            SkillComponentConfiguration, cls._apply_params_to_instance(params, instance)
        )

        return instance

    @classmethod
    def _apply_params_to_instance(
        cls, params: dict, instance: Optional["SkillComponentConfiguration"]
    ) -> "SkillComponentConfiguration":
        """Constructs or update instance with params provided."""
        if instance is None:
            instance = cls(**params)
        else:  # pragma: nocover
            instance.__init__(**params)  # type: ignore
        return instance


class SkillConfig(ComponentConfiguration):
    """Class to represent a skill configuration file."""

    default_configuration_filename = DEFAULT_SKILL_CONFIG_FILE
    package_type = PackageType.SKILL
    schema = "skill-config_schema.json"
    abstract_field_name = "is_abstract"

    FIELDS_ALLOWED_TO_UPDATE: FrozenSet[str] = frozenset(
        ["behaviours", "handlers", "models", "is_abstract", "build_directory"]
    )
    FIELDS_WITH_NESTED_FIELDS: FrozenSet[str] = frozenset(
        ["behaviours", "handlers", "models"]
    )
    NESTED_FIELDS_ALLOWED_TO_UPDATE: FrozenSet[str] = frozenset(["args"])

    __slots__ = (
        "connections",
        "protocols",
        "contracts",
        "skills",
        "dependencies",
        "description",
        "handlers",
        "behaviours",
        "models",
        "is_abstract",
    )

    def __init__(
        self,
        name: SimpleIdOrStr,
        author: SimpleIdOrStr,
        version: str = "",
        license_: str = "",
        aea_version: str = "",
        fingerprint: Optional[Dict[str, str]] = None,
        fingerprint_ignore_patterns: Optional[Sequence[str]] = None,
        build_entrypoint: Optional[str] = None,
        build_directory: Optional[str] = None,
        connections: Optional[Set[PublicId]] = None,
        protocols: Optional[Set[PublicId]] = None,
        contracts: Optional[Set[PublicId]] = None,
        skills: Optional[Set[PublicId]] = None,
        dependencies: Optional[Dependencies] = None,
        description: str = "",
        is_abstract: bool = False,
    ) -> None:
        """Initialize a skill configuration."""
        super().__init__(
            name,
            author,
            version,
            license_,
            aea_version,
            fingerprint,
            fingerprint_ignore_patterns,
            build_entrypoint,
            build_directory,
            dependencies,
        )
        self.connections = connections if connections is not None else set()
        self.protocols = protocols if protocols is not None else set()
        self.contracts = contracts if contracts is not None else set()
        self.skills = skills if skills is not None else set()
        self.dependencies = dependencies if dependencies is not None else {}
        self.description = description
        self.handlers: CRUDCollection[SkillComponentConfiguration] = CRUDCollection()
        self.behaviours: CRUDCollection[SkillComponentConfiguration] = CRUDCollection()
        self.models: CRUDCollection[SkillComponentConfiguration] = CRUDCollection()

        self.is_abstract = is_abstract

    @property
    def package_dependencies(self) -> Set[ComponentId]:
        """Get the skill dependencies."""
        return (
            {
                ComponentId(ComponentType.PROTOCOL, protocol_id)
                for protocol_id in self.protocols
            }
            .union(
                {
                    ComponentId(ComponentType.CONTRACT, contract_id)
                    for contract_id in self.contracts
                }
            )
            .union(
                {ComponentId(ComponentType.SKILL, skill_id) for skill_id in self.skills}
            )
            .union(
                {
                    ComponentId(ComponentType.CONNECTION, connection_id)
                    for connection_id in self.connections
                }
            )
        )

    @property
    def is_abstract_component(self) -> bool:
        """Check whether the component is abstract."""
        return self.is_abstract  # pragma: nocover

    @property
    def json(self) -> Dict:
        """Return the JSON representation."""
        result = OrderedDict(
            {
                "name": self.name,
                "author": self.author,
                "version": self.version,
                "type": self.component_type.value,
                "description": self.description,
                "license": self.license,
                "aea_version": self.aea_version,
                "fingerprint": self.fingerprint,
                "fingerprint_ignore_patterns": self.fingerprint_ignore_patterns,
                CONNECTIONS: sorted(map(str, self.connections)),
                CONTRACTS: sorted(map(str, self.contracts)),
                PROTOCOLS: sorted(map(str, self.protocols)),
                SKILLS: sorted(map(str, self.skills)),
                "behaviours": {key: b.json for key, b in self.behaviours.read_all()},
                "handlers": {key: h.json for key, h in self.handlers.read_all()},
                "models": {key: m.json for key, m in self.models.read_all()},
                "dependencies": dependencies_to_json(self.dependencies),
                "is_abstract": self.is_abstract,
            }
        )
        if self.build_entrypoint:
            result["build_entrypoint"] = self.build_entrypoint
        if self.build_directory:
            result["build_directory"] = self.build_directory
        return result

    @classmethod
    def _create_or_update_from_json(
        cls, obj: Dict, instance: Optional["SkillConfig"] = None
    ) -> "SkillConfig":
        """Initialize from a JSON object."""
        obj = {**(instance.json if instance else {}), **copy(obj)}
        name = cast(str, obj.get("name"))
        author = cast(str, obj.get("author"))
        version = cast(str, obj.get("version"))
        license_ = cast(str, obj.get("license"))
        aea_version_specifiers = cast(str, obj.get("aea_version", ""))
        fingerprint = cast(Dict[str, str], obj.get("fingerprint"))
        fingerprint_ignore_patterns = cast(
            Sequence[str], obj.get("fingerprint_ignore_patterns")
        )
        build_entrypoint = cast(Optional[str], obj.get("build_entrypoint"))
        connections = {PublicId.from_str(id_) for id_ in obj.get(CONNECTIONS, set())}
        protocols = {PublicId.from_str(id_) for id_ in obj.get(PROTOCOLS, set())}
        contracts = {PublicId.from_str(id_) for id_ in obj.get(CONTRACTS, set())}
        skills = {PublicId.from_str(id_) for id_ in obj.get(SKILLS, set())}
        dependencies = dependencies_from_json(obj.get("dependencies", {}))
        description = cast(str, obj.get("description", ""))
        params = dict(
            name=name,
            author=author,
            version=version,
            license_=license_,
            aea_version=aea_version_specifiers,
            fingerprint=fingerprint,
            fingerprint_ignore_patterns=fingerprint_ignore_patterns,
            build_entrypoint=build_entrypoint,
            connections=connections,
            protocols=protocols,
            contracts=contracts,
            skills=skills,
            dependencies=dependencies,
            description=description,
            is_abstract=obj.get("is_abstract", False),
            build_directory=obj.get("build_directory"),
        )

        instance = cast(SkillConfig, cls._apply_params_to_instance(params, instance))

        for behaviour_id, behaviour_data in obj.get("behaviours", {}).items():
            behaviour_config = SkillComponentConfiguration.from_json(behaviour_data)
            instance.behaviours.create(behaviour_id, behaviour_config)

        for handler_id, handler_data in obj.get("handlers", {}).items():
            handler_config = SkillComponentConfiguration.from_json(handler_data)
            instance.handlers.create(handler_id, handler_config)

        for model_id, model_data in obj.get("models", {}).items():
            model_config = SkillComponentConfiguration.from_json(model_data)
            instance.models.create(model_id, model_config)

        return instance

    def get_overridable(self) -> dict:
        """Get overrideable confg data."""
        result = {}
        current_config_data = self.json
        if self.abstract_field_name in current_config_data:
            result[self.abstract_field_name] = current_config_data[
                self.abstract_field_name
            ]

        for field in self.FIELDS_WITH_NESTED_FIELDS:
            if not current_config_data.get(field, {}):
                continue
            result[field] = {}
            for name in current_config_data[field].keys():
                result[field][name] = {}
                for nested_field in self.NESTED_FIELDS_ALLOWED_TO_UPDATE:
                    result[field][name][nested_field] = current_config_data[field][
                        name
                    ][nested_field]
        return result


class AgentConfig(PackageConfiguration):
    """Class to represent the agent configuration file."""

    default_configuration_filename = DEFAULT_AEA_CONFIG_FILE
    package_type = PackageType.AGENT
    schema = "aea-config_schema.json"

    FIELDS_ALLOWED_TO_UPDATE: FrozenSet[str] = frozenset(
        [
            "description",
            "registry_path",
            "logging_config",
            "private_key_paths",
            "connection_private_key_paths",
            "loop_mode",
            "runtime_mode",
            "execution_timeout",
            "timeout",
            "period",
            "max_reactions",
            "skill_exception_policy",
            "connection_exception_policy",
            "default_connection",
            "default_ledger",
            "default_routing",
            "storage_uri",
        ]
    )
    CHECK_EXCLUDES = [
        ("private_key_paths",),
        ("connection_private_key_paths",),
        ("decision_maker_handler",),
        ("default_routing",),
        ("dependencies",),
        ("logging_config",),
    ]

    __slots__ = (
        "agent_name",
        "registry_path",
        "description",
        "private_key_paths",
        "connection_private_key_paths",
        "logging_config",
        "default_ledger",
        "currency_denominations",
        "default_connection",
        "connections",
        "protocols",
        "skills",
        "contracts",
        "period",
        "execution_timeout",
        "max_reactions",
        "skill_exception_policy",
        "connection_exception_policy",
        "error_handler",
        "decision_maker_handler",
        "default_routing",
        "loop_mode",
        "runtime_mode",
        "storage_uri",
        "data_dir",
        "_component_configurations",
        "dependencies",
    )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        agent_name: SimpleIdOrStr,
        author: SimpleIdOrStr,
        version: str = "",
        license_: str = "",
        aea_version: str = "",
        fingerprint: Optional[Dict[str, str]] = None,
        fingerprint_ignore_patterns: Optional[Sequence[str]] = None,
        build_entrypoint: Optional[str] = None,
        registry_path: str = DEFAULT_REGISTRY_NAME,
        description: str = "",
        logging_config: Optional[Dict] = None,
        period: Optional[float] = None,
        execution_timeout: Optional[float] = None,
        max_reactions: Optional[int] = None,
        error_handler: Optional[Dict] = None,
        decision_maker_handler: Optional[Dict] = None,
        skill_exception_policy: Optional[str] = None,
        connection_exception_policy: Optional[str] = None,
        default_ledger: Optional[str] = None,
        currency_denominations: Optional[Dict[str, str]] = None,
        default_connection: Optional[str] = None,
        default_routing: Optional[Dict[str, str]] = None,
        loop_mode: Optional[str] = None,
        runtime_mode: Optional[str] = None,
        storage_uri: Optional[str] = None,
        data_dir: Optional[str] = None,
        component_configurations: Optional[Dict[ComponentId, Dict]] = None,
        dependencies: Optional[Dependencies] = None,
    ) -> None:
        """Instantiate the agent configuration object."""
        super().__init__(
            agent_name,
            author,
            version,
            license_,
            aea_version,
            fingerprint,
            fingerprint_ignore_patterns,
            build_entrypoint,
        )
        self.agent_name = self.name
        self.registry_path = registry_path
        self.description = description
        self.private_key_paths = CRUDCollection[str]()
        self.connection_private_key_paths = CRUDCollection[str]()

        self.logging_config = logging_config or DEFAULT_LOGGING_CONFIG
        self.default_ledger = default_ledger
        self.currency_denominations = (
            currency_denominations if currency_denominations is not None else {}
        )
        self.default_connection = (
            PublicId.from_str(default_connection)
            if default_connection is not None
            else None
        )
        self.connections = set()  # type: Set[PublicId]
        self.contracts = set()  # type: Set[PublicId]
        self.protocols = set()  # type: Set[PublicId]
        self.skills = set()  # type: Set[PublicId]

        self.period: Optional[float] = period
        self.execution_timeout: Optional[float] = execution_timeout
        self.max_reactions: Optional[int] = max_reactions

        self.skill_exception_policy: Optional[str] = skill_exception_policy
        self.connection_exception_policy: Optional[str] = connection_exception_policy

        self.error_handler = error_handler if error_handler is not None else {}
        self.decision_maker_handler = (
            decision_maker_handler if decision_maker_handler is not None else {}
        )

        self.default_routing = (
            {
                PublicId.from_str(key): PublicId.from_str(value)
                for key, value in default_routing.items()
            }
            if default_routing is not None
            else {}
        )  # type: Dict[PublicId, PublicId]
        self.loop_mode = loop_mode
        self.runtime_mode = runtime_mode
        self.storage_uri = storage_uri
        self.data_dir = data_dir
        # this attribute will be set through the setter below
        self._component_configurations: Dict[ComponentId, Dict] = {}
        self.component_configurations = (
            component_configurations if component_configurations is not None else {}
        )
        self.dependencies = dependencies or {}

    @property
    def component_configurations(self) -> Dict[ComponentId, Dict]:
        """Get the custom component configurations."""
        return self._component_configurations

    @component_configurations.setter
    def component_configurations(self, d: Dict[ComponentId, Dict]) -> None:
        """Set the component configurations."""
        package_type_to_set = {
            PackageType.PROTOCOL: self.protocols,
            PackageType.CONNECTION: self.connections,
            PackageType.CONTRACT: self.contracts,
            PackageType.SKILL: self.skills,
        }
        for component_id, component_configuration in d.items():
            enforce(
                component_id.public_id
                in package_type_to_set[component_id.package_type],
                f"Component {component_id} not declared in the agent configuration.",
            )
            ConfigValidator.validate_component_configuration(
                component_id, component_configuration
            )
        self._component_configurations = d

    @property
    def package_dependencies(self) -> Set[ComponentId]:
        """Get the package dependencies."""
        protocols = set(
            ComponentId(ComponentType.PROTOCOL, public_id)
            for public_id in self.protocols
        )
        connections = set(
            ComponentId(ComponentType.CONNECTION, public_id)
            for public_id in self.connections
        )
        skills = set(
            ComponentId(ComponentType.SKILL, public_id) for public_id in self.skills
        )

        contracts = set(
            ComponentId(ComponentType.CONTRACT, public_id)
            for public_id in self.contracts
        )

        return set.union(protocols, contracts, connections, skills)

    @property
    def private_key_paths_dict(self) -> Dict[str, str]:
        """Get dictionary version of private key paths."""
        return {  # pylint: disable=unnecessary-comprehension
            key: path for key, path in self.private_key_paths.read_all()
        }

    @property
    def connection_private_key_paths_dict(self) -> Dict[str, str]:
        """Get dictionary version of connection private key paths."""
        return {  # pylint: disable=unnecessary-comprehension
            key: path for key, path in self.connection_private_key_paths.read_all()
        }

    def component_configurations_json(self) -> List[OrderedDict]:
        """Get the component configurations in JSON format."""
        result: List[OrderedDict] = []
        for component_id, config in self.component_configurations.items():
            result.append(
                OrderedDict(
                    public_id=str(component_id.public_id),
                    type=str(component_id.component_type),
                    **config,
                )
            )
        return result

    @property
    def json(self) -> Dict:
        """Return the JSON representation."""
        config = OrderedDict(
            {
                "agent_name": self.agent_name,
                "author": self.author,
                "version": self.version,
                "license": self.license,
                "description": self.description,
                "aea_version": self.aea_version,
                "fingerprint": self.fingerprint,
                "fingerprint_ignore_patterns": self.fingerprint_ignore_patterns,
                CONNECTIONS: sorted(map(str, self.connections)),
                CONTRACTS: sorted(map(str, self.contracts)),
                PROTOCOLS: sorted(map(str, self.protocols)),
                SKILLS: sorted(map(str, self.skills)),
                "default_connection": str(self.default_connection)
                if self.default_connection is not None
                else None,
                "default_ledger": self.default_ledger,
                "default_routing": {
                    str(key): str(value) for key, value in self.default_routing.items()
                },
                "connection_private_key_paths": self.connection_private_key_paths_dict,
                "private_key_paths": self.private_key_paths_dict,
                "logging_config": self.logging_config,
                "registry_path": self.registry_path,
                "component_configurations": self.component_configurations_json(),
                "dependencies": dependencies_to_json(self.dependencies),
            }
        )  # type: Dict[str, Any]

        if self.build_entrypoint:
            config["build_entrypoint"] = self.build_entrypoint

        # framework optional configs are only printed if defined.
        if self.period is not None:
            config["period"] = self.period
        if self.execution_timeout is not None:
            config["execution_timeout"] = self.execution_timeout
        if self.max_reactions is not None:
            config["max_reactions"] = self.max_reactions
        if self.error_handler != {}:
            config["error_handler"] = self.error_handler
        if self.decision_maker_handler != {}:
            config["decision_maker_handler"] = self.decision_maker_handler
        if self.skill_exception_policy is not None:
            config["skill_exception_policy"] = self.skill_exception_policy
        if self.connection_exception_policy is not None:
            config["connection_exception_policy"] = self.connection_exception_policy
        if self.loop_mode is not None:
            config["loop_mode"] = self.loop_mode
        if self.runtime_mode is not None:
            config["runtime_mode"] = self.runtime_mode
        if self.storage_uri is not None:
            config["storage_uri"] = self.storage_uri
        if self.data_dir is not None:
            config["data_dir"] = self.data_dir
        if self.currency_denominations != {}:
            config["currency_denominations"] = self.currency_denominations

        return config

    @classmethod
    def _create_or_update_from_json(
        cls, obj: Dict, instance: Optional[Any] = None
    ) -> "AgentConfig":
        """Create new config object or updates existing one from json data."""
        obj = {**(instance.json if instance else {}), **copy(obj)}
        params = dict(
            agent_name=cast(str, obj.get("agent_name")),
            author=cast(str, obj.get("author")),
            version=cast(str, obj.get("version")),
            license_=cast(str, obj.get("license")),
            aea_version=cast(str, obj.get("aea_version", "")),
            registry_path=cast(str, obj.get("registry_path")),
            description=cast(str, obj.get("description", "")),
            fingerprint=cast(Dict[str, str], obj.get("fingerprint", {})),
            fingerprint_ignore_patterns=cast(
                Sequence[str], obj.get("fingerprint_ignore_patterns")
            ),
            build_entrypoint=cast(Optional[str], obj.get("build_entrypoint")),
            logging_config=cast(Dict, obj.get("logging_config", {})),
            period=cast(float, obj.get("period")),
            execution_timeout=cast(float, obj.get("execution_timeout")),
            max_reactions=cast(int, obj.get("max_reactions")),
            error_handler=cast(Dict, obj.get("error_handler", {})),
            decision_maker_handler=cast(Dict, obj.get("decision_maker_handler", {})),
            skill_exception_policy=cast(str, obj.get("skill_exception_policy")),
            connection_exception_policy=cast(
                str, obj.get("connection_exception_policy")
            ),
            default_ledger=cast(str, obj.get("default_ledger")),
            currency_denominations=cast(Dict, obj.get("currency_denominations", {})),
            default_connection=cast(str, obj.get("default_connection")),
            default_routing=cast(Dict, obj.get("default_routing", {})),
            loop_mode=cast(str, obj.get("loop_mode")),
            runtime_mode=cast(str, obj.get("runtime_mode")),
            storage_uri=cast(str, obj.get("storage_uri")),
            data_dir=cast(str, obj.get("data_dir")),
            component_configurations=None,
            dependencies=cast(
                Dependencies, dependencies_from_json(obj.get("dependencies", {}))
            ),
        )
        instance = cast(AgentConfig, cls._apply_params_to_instance(params, instance))

        agent_config = instance

        # Parse private keys
        for crypto_id, path in obj.get("private_key_paths", {}).items():
            agent_config.private_key_paths.create(crypto_id, path)

        for crypto_id, path in obj.get("connection_private_key_paths", {}).items():
            agent_config.connection_private_key_paths.create(crypto_id, path)

        # parse connection public ids
        agent_config.connections = set(
            map(PublicId.from_str, obj.get(CONNECTIONS, []),)
        )

        # parse contracts public ids
        agent_config.contracts = set(map(PublicId.from_str, obj.get(CONTRACTS, []),))

        # parse protocol public ids
        agent_config.protocols = set(map(PublicId.from_str, obj.get(PROTOCOLS, []),))

        # parse skills public ids
        agent_config.skills = set(map(PublicId.from_str, obj.get(SKILLS, []),))

        # parse component configurations
        component_configurations = {}
        for config in obj.get("component_configurations", []):
            tmp = deepcopy(config)
            public_id = PublicId.from_str(tmp.pop("public_id"))
            type_ = tmp.pop("type")
            component_id = ComponentId(ComponentType(type_), public_id)
            component_configurations[component_id] = tmp
        agent_config.component_configurations = component_configurations

        return agent_config

    @property
    def all_components_id(self) -> List[ComponentId]:
        """Get list of the all components for this agent config."""
        component_type_to_set = {
            ComponentType.PROTOCOL: self.protocols,
            ComponentType.CONNECTION: self.connections,
            ComponentType.CONTRACT: self.contracts,
            ComponentType.SKILL: self.skills,
        }
        result = []
        for component_type, public_ids in component_type_to_set.items():
            for public_id in public_ids:
                result.append(ComponentId(component_type, public_id))

        return result

    def update(self, data: Dict, env_vars_friendly: bool = False) -> None:
        """
        Update configuration with other data.

        To update the component parts, populate the field "component_configurations" as a
        mapping from ComponentId to configurations.

        :param data: the data to replace.
        :return: None
        """
        data = copy(data)
        # update component parts
        new_component_configurations: Dict = data.pop("component_configurations", {})
        updated_component_configurations: Dict[ComponentId, Dict] = copy(
            self.component_configurations
        )
        for component_id, obj in new_component_configurations.items():
            if component_id not in updated_component_configurations:
                updated_component_configurations[component_id] = obj
            else:
                recursive_update(
                    updated_component_configurations[component_id],
                    obj,
                    allow_new_values=True,
                )

        self.check_overrides_valid(data, env_vars_friendly=env_vars_friendly)
        super().update(data, env_vars_friendly=env_vars_friendly)
        self.validate_config_data(self.json, env_vars_friendly=env_vars_friendly)
        self.component_configurations = updated_component_configurations


class SpeechActContentConfig(Configuration):
    """Handle a speech_act content configuration."""

    __slots__ = ("args",)

    def __init__(self, **args: Any) -> None:
        """Initialize a speech_act content configuration."""
        super().__init__()
        self.args = args  # type: Dict[str, str]

    @property
    def json(self) -> Dict:
        """Return the JSON representation."""
        return self.args

    @classmethod
    def from_json(cls, obj: Dict) -> "SpeechActContentConfig":
        """Initialize from a JSON object."""
        return SpeechActContentConfig(**obj)


class ProtocolSpecification(ProtocolConfig):
    """Handle protocol specification."""

    __slots__ = ("speech_acts", "_protobuf_snippets", "_dialogue_config")

    def __init__(
        self,
        name: SimpleIdOrStr,
        author: SimpleIdOrStr,
        version: str = "",
        license_: str = "",
        aea_version: str = "",
        description: str = "",
        protocol_specification_id: Optional[str] = None,
    ) -> None:
        """Initialize a protocol specification configuration object."""
        super().__init__(
            name,
            author,
            version,
            license_,
            aea_version=aea_version,
            description=description,
            protocol_specification_id=protocol_specification_id,
        )
        self.speech_acts = CRUDCollection[SpeechActContentConfig]()
        self._protobuf_snippets = {}  # type: Dict
        self._dialogue_config = {}  # type: Dict

    @property
    def protobuf_snippets(self) -> Dict:
        """Get the protobuf snippets."""
        return self._protobuf_snippets

    @protobuf_snippets.setter
    def protobuf_snippets(self, protobuf_snippets: Dict) -> None:
        """Set the protobuf snippets."""
        self._protobuf_snippets = protobuf_snippets

    @property
    def dialogue_config(self) -> Dict:
        """Get the dialogue config."""
        return self._dialogue_config

    @dialogue_config.setter
    def dialogue_config(self, dialogue_config: Dict) -> None:
        """Set the dialogue config."""
        self._dialogue_config = dialogue_config

    @property
    def json(self) -> Dict:
        """Return the JSON representation."""
        result: Dict[str, Any] = OrderedDict(
            {
                "name": self.name,
                "author": self.author,
                "version": self.version,
                "description": self.description,
                "license": self.license,
                "aea_version": self.aea_version,
                "protocol_specification_id": str(self.protocol_specification_id),
                "speech_acts": {
                    key: speech_act.json
                    for key, speech_act in self.speech_acts.read_all()
                },
            }
        )
        return result

    @classmethod
    def _create_or_update_from_json(  # type: ignore
        cls, obj: Dict, instance: Optional["ProtocolSpecification"] = None
    ) -> "ProtocolSpecification":
        """Initialize from a JSON object."""
        obj = {**(instance.json if instance else {}), **copy(obj)}
        params = dict(
            name=cast(str, obj.get("name")),
            author=cast(str, obj.get("author")),
            protocol_specification_id=cast(str, obj.get("protocol_specification_id")),
            version=cast(str, obj.get("version")),
            license_=cast(str, obj.get("license")),
            aea_version=cast(str, obj.get("aea_version", "")),
            description=cast(str, obj.get("description", "")),
        )

        instance = cast(
            ProtocolSpecification, cls._apply_params_to_instance(params, instance)
        )

        protocol_specification = instance
        for speech_act, speech_act_content in obj.get("speech_acts", {}).items():
            speech_act_content_config = SpeechActContentConfig.from_json(
                speech_act_content
            )
            protocol_specification.speech_acts.create(
                speech_act, speech_act_content_config
            )
        return protocol_specification


class ContractConfig(ComponentConfiguration):
    """Handle contract configuration."""

    default_configuration_filename = DEFAULT_CONTRACT_CONFIG_FILE
    package_type = PackageType.CONTRACT
    schema = "contract-config_schema.json"

    FIELDS_ALLOWED_TO_UPDATE: FrozenSet[str] = frozenset(["build_directory"])

    __slots__ = (
        "dependencies",
        "description",
        "contract_interface_paths",
        "class_name",
    )

    def __init__(
        self,
        name: SimpleIdOrStr,
        author: SimpleIdOrStr,
        version: str = "",
        license_: str = "",
        aea_version: str = "",
        fingerprint: Optional[Dict[str, str]] = None,
        fingerprint_ignore_patterns: Optional[Sequence[str]] = None,
        build_entrypoint: Optional[str] = None,
        build_directory: Optional[str] = None,
        dependencies: Optional[Dependencies] = None,
        description: str = "",
        contract_interface_paths: Optional[Dict[str, str]] = None,
        class_name: str = "",
    ) -> None:
        """Initialize a protocol configuration object."""
        super().__init__(
            name,
            author,
            version,
            license_,
            aea_version,
            fingerprint,
            fingerprint_ignore_patterns,
            build_entrypoint,
            build_directory,
            dependencies,
        )
        self.dependencies = dependencies if dependencies is not None else {}
        self.description = description
        self.contract_interface_paths = (
            contract_interface_paths if contract_interface_paths is not None else {}
        )
        self.class_name = class_name

    @property
    def json(self) -> Dict:
        """Return the JSON representation."""
        result = OrderedDict(
            {
                "name": self.name,
                "author": self.author,
                "version": self.version,
                "type": self.component_type.value,
                "description": self.description,
                "license": self.license,
                "aea_version": self.aea_version,
                "fingerprint": self.fingerprint,
                "fingerprint_ignore_patterns": self.fingerprint_ignore_patterns,
                "class_name": self.class_name,
                "contract_interface_paths": self.contract_interface_paths,
                "dependencies": dependencies_to_json(self.dependencies),
            }
        )
        if self.build_entrypoint:
            result["build_entrypoint"] = self.build_entrypoint
        if self.build_directory:
            result["build_directory"] = self.build_directory
        return result

    @classmethod
    def _create_or_update_from_json(
        cls, obj: Dict, instance: Optional["ContractConfig"] = None
    ) -> "ContractConfig":
        """Initialize from a JSON object."""
        obj = {**(instance.json if instance else {}), **copy(obj)}
        dependencies = cast(
            Dependencies, dependencies_from_json(obj.get("dependencies", {}))
        )
        params = dict(
            name=cast(str, obj.get("name")),
            author=cast(str, obj.get("author")),
            version=cast(str, obj.get("version")),
            license_=cast(str, obj.get("license")),
            aea_version=cast(str, obj.get("aea_version", "")),
            fingerprint=cast(Dict[str, str], obj.get("fingerprint", {})),
            fingerprint_ignore_patterns=cast(
                Sequence[str], obj.get("fingerprint_ignore_patterns")
            ),
            build_entrypoint=cast(Optional[str], obj.get("build_entrypoint")),
            build_directory=cast(Optional[str], obj.get("build_directory")),
            dependencies=dependencies,
            description=cast(str, obj.get("description", "")),
            contract_interface_paths=cast(
                Dict[str, str], obj.get("contract_interface_paths", {})
            ),
            class_name=obj.get("class_name", ""),
        )
        instance = cast(ContractConfig, cls._apply_params_to_instance(params, instance))

        return instance


"""The following functions are called from aea.cli.utils."""


def _compute_fingerprint(
    package_directory: Path, ignore_patterns: Optional[Collection[str]] = None
) -> Dict[str, str]:
    ignore_patterns = ignore_patterns if ignore_patterns is not None else []
    ignore_patterns = set(ignore_patterns).union(DEFAULT_FINGERPRINT_IGNORE_PATTERNS)
    hasher = IPFSHashOnly()
    fingerprints = {}  # type: Dict[str, str]
    # find all valid files of the package
    all_files = [
        x
        for x in package_directory.glob("**/*")
        if x.is_file()
        and (
            x.match("*.py") or not any(x.match(pattern) for pattern in ignore_patterns)
        )
    ]

    for file in all_files:
        file_hash = hasher.get(str(file))
        key = str(file.relative_to(package_directory))
        enforce(key not in fingerprints, "Key in fingerprints!")  # nosec
        # use '/' as path separator
        normalized_path = Path(key).as_posix()
        fingerprints[normalized_path] = file_hash

    return fingerprints


def _compare_fingerprints(
    package_configuration: PackageConfiguration,
    package_directory: Path,
    is_vendor: bool,
    item_type: PackageType,
) -> None:
    """
    Check fingerprints of a package directory against the fingerprints declared in the configuration file.

    :param package_configuration: the package configuration object.
    :param package_directory: the directory of the package.
    :param is_vendor: whether the package is vendorized or not.
    :param item_type: the type of the item.
    :return: None
    :raises ValueError: if the fingerprints do not match.
    """
    expected_fingerprints = package_configuration.fingerprint
    ignore_patterns = package_configuration.fingerprint_ignore_patterns
    actual_fingerprints = _compute_fingerprint(package_directory, ignore_patterns)
    if expected_fingerprints != actual_fingerprints:
        if is_vendor:
            raise ValueError(
                (
                    "Fingerprints for package {} do not match:\nExpected: {}\nActual: {}\n"
                    "Vendorized projects should not be tampered with, please revert any changes to {} {}"
                ).format(
                    package_directory,
                    pprint.pformat(expected_fingerprints),
                    pprint.pformat(actual_fingerprints),
                    str(item_type),
                    package_configuration.public_id,
                )
            )
        raise ValueError(
            (
                "Fingerprints for package {} do not match:\nExpected: {}\nActual: {}\n"
                "Please fingerprint the package before continuing: 'aea fingerprint {} {}'"
            ).format(
                package_directory,
                pprint.pformat(expected_fingerprints),
                pprint.pformat(actual_fingerprints),
                str(item_type),
                package_configuration.public_id,
            )
        )


class AEAVersionError(ValueError):
    """Special Exception for version error."""

    def __init__(
        self, package_id: PublicId, aea_version_specifiers: SpecifierSet
    ) -> None:
        """Init exception."""
        self.package_id = package_id
        self.aea_version_specifiers = aea_version_specifiers
        self.current_aea_version = Version(__aea_version__)
        super().__init__(
            f"The CLI version is {self.current_aea_version}, but package {self.package_id} requires version {self.aea_version_specifiers}"
        )


def _check_aea_version(package_configuration: PackageConfiguration) -> None:
    """Check the package configuration version against the version of the framework."""
    current_aea_version = Version(__aea_version__)
    version_specifiers = package_configuration.aea_version_specifiers
    if current_aea_version not in version_specifiers:
        raise AEAVersionError(
            package_configuration.public_id,
            package_configuration.aea_version_specifiers,
        )


def _compare_public_ids(
    component_configuration: ComponentConfiguration, package_directory: Path
) -> None:
    """Compare the public ids in config and init file."""
    if component_configuration.package_type != PackageType.SKILL:
        return
    filename = "__init__.py"
    public_id_in_init = _get_public_id_from_file(
        component_configuration, package_directory, filename
    )
    if (
        public_id_in_init is not None
        and public_id_in_init != component_configuration.public_id
    ):
        raise ValueError(  # pragma: nocover
            f"The public id specified in {filename} for package {package_directory} does not match the one specific in {component_configuration.package_type.value}.yaml"
        )


def _get_public_id_from_file(
    component_configuration: ComponentConfiguration,
    package_directory: Path,
    filename: str,
) -> Optional[PublicId]:
    """
    Get the public id from an init if present.

    :param component_configuration: the component configuration.
    :param package_directory: the path to the package directory.
    :param filename: the file
    :return: the public id, if found.
    """
    path_to_file = Path(package_directory, filename)
    module = load_module(component_configuration.prefix_import_path, path_to_file)
    package_public_id = getattr(module, PACKAGE_PUBLIC_ID_VAR_NAME, None)
    return package_public_id


PACKAGE_TYPE_TO_CONFIG_CLASS: Dict[PackageType, Type[PackageConfiguration]] = {
    PackageType.AGENT: AgentConfig,
    PackageType.PROTOCOL: ProtocolConfig,
    PackageType.CONNECTION: ConnectionConfig,
    PackageType.SKILL: SkillConfig,
    PackageType.CONTRACT: ContractConfig,
}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the content of generic-storage.md file."""
import inspect
import re
from pathlib import Path

from aea.helpers.storage.generic_storage import SyncCollection
from aea.skills.behaviours import TickerBehaviour

from tests.conftest import ROOT_DIR
from tests.test_docs.helper import BaseTestMarkdownDocs


class TestGenericStorage(BaseTestMarkdownDocs):
    """Test generic-storage.md documentation."""

    DOC_PATH = Path(ROOT_DIR, "docs", "generic-storage.md")

    @classmethod
    def setup_class(cls):
        """
        Set up test.

        We test only the signatures and the
        docstrings of 'SyncCollection' methods. Hence,
        we need to remove the return statement.
        """
        super().setup_class()
        cls.content = inspect.getsource(SyncCollection)
        cls.content = re.sub("\n +return.*", "", cls.content)

    def test_storage_abstract_methods(self):
        """
        Test storage abstract methods.

        Note: the block index of the abstract methods is 0.
        Please check generic-storage.md
        """
        block_index = 1
        assert self.code_blocks[block_index]["text"] in self.content

    def test_test_behaviour(self):
        """Test that the 'TestBehaviour' code is compilable."""
        block_index = 2
        code = self.code_blocks[block_index]["text"]
        exec(code, {}, dict(TickerBehaviour=TickerBehaviour))  # nosec



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Prometheus connection and channel."""

import asyncio
import logging
from typing import Any, Dict, Optional, Tuple, Union, cast

import aioprometheus  # type: ignore

from aea.common import Address
from aea.configurations.base import PublicId
from aea.connections.base import Connection, ConnectionStates
from aea.exceptions import enforce
from aea.mail.base import Envelope, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue

from packages.fetchai.protocols.prometheus.dialogues import PrometheusDialogue
from packages.fetchai.protocols.prometheus.dialogues import (
    PrometheusDialogues as BasePrometheusDialogues,
)
from packages.fetchai.protocols.prometheus.message import PrometheusMessage


PUBLIC_ID = PublicId.from_str("fetchai/prometheus:0.4.0")

DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 9090
VALID_UPDATE_FUNCS = {"inc", "dec", "add", "sub", "set", "observe"}
VALID_METRIC_TYPES = {"Counter", "Gauge", "Histogram", "Summary"}


class PrometheusDialogues(BasePrometheusDialogues):
    """The dialogues class keeps track of all prometheus dialogues."""

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            # The server connection maintains the dialogue on behalf of the agent
            return PrometheusDialogue.Role.SERVER

        BasePrometheusDialogues.__init__(
            self,
            self_address=str(PUBLIC_ID),
            role_from_first_message=role_from_first_message,
            **kwargs,
        )


class PrometheusChannel:
    """A wrapper for interacting with a prometheus server."""

    def __init__(
        self,
        address: Address,
        host: str,
        port: int,
        logger: Union[logging.Logger, logging.LoggerAdapter],
    ):
        """
        Initialize a prometheus channel.

        :param address: The address of the connection.
        :param metrics: The prometheus metrics.
        :param port: The port at which to expose the metrics.
        """
        self.address = address
        self.metrics = {}  # type: Dict[str, aioprometheus.Collector]
        self.logger = logger
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._queue: Optional[asyncio.Queue] = None
        self._dialogues = PrometheusDialogues()
        self._host = host
        self._port = port
        self._service = aioprometheus.Service()

    def _get_message_and_dialogue(
        self, envelope: Envelope
    ) -> Tuple[PrometheusMessage, Optional[PrometheusDialogue]]:
        """
        Get a message copy and dialogue related to this message.

        :param envelope: incoming envelope

        :return: Tuple[Message, Optional[Dialogue]]
        """
        message = cast(PrometheusMessage, envelope.message)
        dialogue = cast(Optional[PrometheusDialogue], self._dialogues.update(message))
        return message, dialogue

    @property
    def queue(self) -> asyncio.Queue:
        """Check queue is set and return queue."""
        if self._queue is None:  # pragma: nocover
            raise ValueError("Channel is not connected")
        return self._queue

    async def connect(self) -> None:
        """
        Start prometheus http server.

        :return: None
        """
        if self._queue:  # pragma: nocover
            return None
        self._loop = asyncio.get_event_loop()
        self._queue = asyncio.Queue()
        await self._service.start(addr=self._host, port=self._port)

    async def send(self, envelope: Envelope) -> None:
        """
        Process the envelopes to prometheus.

        :return: None
        """
        sender = envelope.sender
        self.logger.debug("Processing message from {}: {}".format(sender, envelope))
        if (
            envelope.protocol_specification_id
            != PrometheusMessage.protocol_specification_id
        ):
            raise ValueError(
                f"Protocol {envelope.protocol_specification_id} is not valid for prometheus."
            )
        await self._handle_prometheus_message(envelope)

    async def _handle_prometheus_message(self, envelope: Envelope) -> None:
        """
        Handle messages to prometheus.

        :param envelope: the envelope
        :return: None
        """
        enforce(
            isinstance(envelope.message, PrometheusMessage),
            "Message not of type PrometheusMessage",
        )
        message, dialogue = self._get_message_and_dialogue(envelope)

        if dialogue is None:
            self.logger.warning(
                "Could not create dialogue from message={}".format(message)
            )
            return

        if message.performative == PrometheusMessage.Performative.ADD_METRIC:
            response = await self._handle_add_metric(message)
        elif message.performative == PrometheusMessage.Performative.UPDATE_METRIC:
            response = await self._handle_update_metric(message)
        else:  # pragma: nocover
            self.logger.warning("Unrecognized performative for PrometheusMessage")
            return

        response_code, response_msg = cast(Tuple[int, str], response)

        msg = dialogue.reply(
            performative=PrometheusMessage.Performative.RESPONSE,
            target_message=message,
            code=response_code,
            message=response_msg,
        )
        envelope = Envelope(to=msg.to, sender=msg.sender, message=msg)
        await self._send(envelope)

    async def _handle_add_metric(self, message: PrometheusMessage) -> Tuple[int, str]:
        """Handle add metric message.

        :param message: the message to handle.
        :return: the response code and response message.
        """
        if message.title in self.metrics:
            response_code = 409
            response_msg = "Metric already exists."
        else:
            metric_type = getattr(aioprometheus, message.type, None)
            if metric_type is None or message.type not in VALID_METRIC_TYPES:
                response_code = 404
                response_msg = f"{message.type} is not a recognized prometheus metric."
            else:
                self.metrics[message.title] = metric_type(
                    message.title, message.description, message.labels
                )
                self._service.register(self.metrics[message.title])
                response_code = 200
                response_msg = (
                    f"New {message.type} successfully added: {message.title}."
                )

        return response_code, response_msg

    async def _handle_update_metric(
        self, message: PrometheusMessage
    ) -> Tuple[int, str]:
        """Handle update metric message.

        :param message: the message to handle.
        :return: the response code and response message.
        """
        metric = message.title
        if metric not in self.metrics:
            response_code = 404
            response_msg = f"Metric {metric} not found."
        else:
            update_func = getattr(self.metrics[metric], message.callable, None)
            if update_func is None:
                response_code = 400
                response_msg = (
                    f"Update function {message.callable} not found for metric {metric}."
                )
            else:
                if message.callable in VALID_UPDATE_FUNCS:
                    # Update the metric ("inc" and "dec" do not take "value" argument)
                    if message.callable in {"inc", "dec"}:
                        update_func(message.labels)
                    else:
                        update_func(message.labels, message.value)
                    response_code = 200
                    response_msg = f"Metric {metric} successfully updated."
                else:
                    response_code = 400
                    response_msg = f"Failed to update metric {metric}: {message.callable} is not a valid update function."

        return response_code, response_msg

    async def _send(self, envelope: Envelope) -> None:
        """Send a message.

        :param envelope: the envelope
        :return: None
        """
        await self.queue.put(envelope)

    async def disconnect(self) -> None:
        """
        Disconnect.

        :return: None
        """
        if self._queue is not None:
            await self._queue.put(None)
            self._queue = None
        await self._service.stop()

    async def get(self) -> Optional[Envelope]:
        """Get incoming envelope."""
        return await self.queue.get()


class PrometheusConnection(Connection):
    """Proxy to the functionality of prometheus."""

    connection_id = PUBLIC_ID

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize a connection to a local prometheus server.

        :param kwargs: the keyword arguments of the parent class.
        """
        super().__init__(**kwargs)

        self.host = cast(str, self.configuration.config.get("host", DEFAULT_HOST))
        self.port = cast(int, self.configuration.config.get("port", DEFAULT_PORT))
        self.channel = PrometheusChannel(
            self.address, self.host, self.port, self.logger
        )

    async def connect(self) -> None:
        """
        Connect to prometheus server via prometheus channel.

        :return: None
        """
        if self.is_connected:  # pragma: nocover
            return

        with self._connect_context():
            self.channel.logger = self.logger
            self.state = ConnectionStates.connecting
            await self.channel.connect()
            self.state = ConnectionStates.connected

    async def disconnect(self) -> None:
        """
        Disconnect from prometheus server.

        :return: None
        """
        if self.is_disconnected:  # pragma: nocover
            return

        self.state = ConnectionStates.disconnecting
        await self.channel.disconnect()
        self.state = ConnectionStates.disconnected

    async def send(self, envelope: Envelope) -> None:
        """
        Send an envelope.

        :param envelope: the envelop
        :return: None
        """
        self._ensure_connected()
        await self.channel.send(envelope)

    async def receive(self, *args: Any, **kwargs: Any) -> Optional["Envelope"]:
        """
        Receive an envelope.

        :return: The received envelope or None
        """
        self._ensure_connected()
        try:
            envelope = await self.channel.get()
            return envelope
        except asyncio.CancelledError:  # pragma: no cover
            return None



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Implementation of the 'aea sync-local-registry' subcommand."""
import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator, Tuple, Union, cast

import click

from aea.cli.fingerprint import determine_package_type_for_directory
from aea.cli.registry.add import fetch_package
from aea.cli.registry.utils import get_package_meta
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import pass_ctx
from aea.cli.utils.loggers import logger
from aea.configurations.data_types import PackageId, PackageType, PublicId
from aea.configurations.loader import load_package_configuration


PACKAGES_DIRS = [i.to_plural() for i in PackageType]


@click.command()
@pass_ctx
def local_registry_sync(ctx: Context) -> None:
    """Upgrade the local package registry."""
    skip_consistency_check = ctx.config["skip_consistency_check"]
    do_local_registry_update(ctx.cwd, skip_consistency_check)


def do_local_registry_update(
    base_dir: Union[str, Path], skip_consistency_check: bool = True
) -> None:
    """
    Perform local registry update.

    :param base_dir: root directory of the local registry.

    :return: None
    """
    for package_id, package_dir in enlist_packages(base_dir, skip_consistency_check):
        current_public_id = package_id.public_id
        latest_public_id = get_package_latest_public_id(package_id)
        if not (  # pylint: disable=superfluous-parens
            current_public_id < latest_public_id
        ):
            click.echo(f"{package_id} is the latest version.")
            continue
        click.echo(f"Updating {current_public_id} to {latest_public_id}.")
        replace_package(str(package_id.package_type), latest_public_id, package_dir)


def replace_package(
    package_type: str, public_id: PublicId, package_dir: Union[Path, str]
) -> None:
    """
    Download, extract and replace exists package.

    :param package_type: str.
    :param public_id: pacakge bulic id to download
    :param: package_dir: target package dir

    :return: None
    """
    with TemporaryDirectory() as tmp_dir:
        new_package_dir = os.path.join(tmp_dir, public_id.name)
        os.mkdir(new_package_dir)
        fetch_package(
            package_type, public_id=public_id, cwd=tmp_dir, dest=new_package_dir
        )
        shutil.rmtree(package_dir)
        shutil.move(new_package_dir, package_dir)


def get_package_latest_public_id(package_id: PackageId) -> PublicId:
    """
    Get package latest package id from the remote repo.

    :param package_id: id of the package to check

    :return: package id of the latest package in remote repo
    """
    package_meta = get_package_meta(
        str(package_id.package_type), package_id.public_id.to_latest()
    )
    return PublicId.from_str(cast(str, package_meta["public_id"]))


def enlist_packages(
    base_dir: Union[Path, str], skip_consistency_check: bool = True
) -> Generator[Tuple[PackageId, Union[Path, str]], None, None]:
    """
    Generate list of the packages in local repo directory.

    :param base_dir: path or str of the local repo.

    :return: generator with Tuple of package_id and package diurectory.
    """
    for author in os.listdir(base_dir):
        author_dir = os.path.join(base_dir, author)
        if not os.path.isdir(author_dir):
            continue  # pragma: nocover
        for package_type_plural in os.listdir(author_dir):
            if package_type_plural not in PACKAGES_DIRS:
                continue  # pragma: nocover
            package_type_dir = os.path.join(author_dir, package_type_plural)
            if not os.path.isdir(package_type_dir):
                continue  # pragma: nocover
            for package_name in os.listdir(package_type_dir):
                package_dir = os.path.join(package_type_dir, package_name)
                if not os.path.isdir(package_dir):
                    continue  # pragma: nocover
                try:
                    package_type = determine_package_type_for_directory(
                        Path(package_dir)
                    )
                    if package_type.to_plural() != package_type_plural:
                        # incorrect package placing
                        continue  # pragma: nocover
                    config = load_package_configuration(
                        package_type,
                        Path(package_dir),
                        skip_consistency_check=skip_consistency_check,
                    )
                    yield (config.package_id, package_dir)
                except ValueError as e:  # pragma: nocover
                    logger.error(  # pragma: nocover
                        f"Error with package_dir={package_dir}: {e}"
                    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of plug-in mechanism for cryptos."""
import itertools
import pprint
from typing import Iterator, List, Set

from pkg_resources import EntryPoint, iter_entry_points

from aea.configurations.constants import (
    ALLOWED_GROUPS,
    CRYPTO_PLUGIN_GROUP,
    DOTTED_PATH_MODULE_ELEMENT_SEPARATOR,
    FAUCET_APIS_PLUGIN_GROUP,
    LEDGER_APIS_PLUGIN_GROUP,
)
from aea.crypto import register_crypto, register_faucet_api, register_ledger_api
from aea.crypto.registries.base import EntryPoint as EntryPointString
from aea.crypto.registries.base import ItemId
from aea.exceptions import AEAPluginError, enforce


_from_group_to_register_callable = {
    CRYPTO_PLUGIN_GROUP: register_crypto,
    LEDGER_APIS_PLUGIN_GROUP: register_ledger_api,
    FAUCET_APIS_PLUGIN_GROUP: register_faucet_api,
}


class Plugin:
    """Class that implements an AEA plugin."""

    __slots__ = ("_group", "_entry_point")

    def __init__(self, group: str, entry_point: EntryPoint):
        """
        Initialize the plugin.

        :param group: the group the plugin belongs to.
        :param entry_point: the entrypoint.
        """
        self._group = group
        self._entry_point = entry_point
        self._check_consistency()

    def _check_consistency(self) -> None:
        """
        Check consistency of input.

        :raises AEAPluginError: if some input is not correct.
        """
        _error_message_prefix = f"Error with plugin '{self._entry_point.name}':"
        enforce(
            self.group in ALLOWED_GROUPS,
            f"{_error_message_prefix} '{self.group}' is not in the allowed groups: {pprint.pformat(ALLOWED_GROUPS)}",
            AEAPluginError,
        )
        enforce(
            ItemId.REGEX.match(self._entry_point.name) is not None,
            f"{_error_message_prefix} '{self._entry_point.name}' is not a valid identifier for a plugin.",
            AEAPluginError,
        )
        enforce(
            len(self._entry_point.attrs) == 1,
            f"{_error_message_prefix} Nested attributes currently not supported.",
            AEAPluginError,
        )
        enforce(
            len(self._entry_point.extras) == 0,
            f"{_error_message_prefix} Extras currently not supported.",
            AEAPluginError,
        )
        enforce(
            EntryPointString.REGEX.match(self.entry_point_path) is not None,
            f"{_error_message_prefix} Entry point path '{self.entry_point_path}' is not valid.",
        )

    @property
    def name(self) -> str:
        """Get the plugin identifier."""
        return self._entry_point.name

    @property
    def group(self) -> str:
        """Get the group."""
        return self._group

    @property
    def attr(self) -> str:
        """Get the class name."""
        return self._entry_point.attrs[0]

    @property
    def entry_point_path(self) -> str:
        """Get the entry point path."""
        class_name = self.attr
        return f"{self._entry_point.module_name}{DOTTED_PATH_MODULE_ELEMENT_SEPARATOR}{class_name}"


def _check_no_duplicates(plugins: List[EntryPoint]) -> None:
    """Check there are no two plugins with the same id."""
    seen: Set[str] = set()
    duplicate_plugins = [p for p in plugins if p.name in seen or seen.add(p.name)]  # type: ignore
    error_msg = f"Found plugins with the same id: {pprint.pformat(duplicate_plugins)}"
    enforce(len(duplicate_plugins) == 0, error_msg, AEAPluginError)


def _get_plugins(group: str) -> List[Plugin]:
    """
    Return a dict of all installed plugins, by name.

    :param group: the plugin group.
    :return: a mapping from plugin name to Plugin objects.
    """
    entry_points: List[EntryPoint] = list(iter_entry_points(group=group))
    _check_no_duplicates(entry_points)
    return [Plugin(group, entry_point) for entry_point in entry_points]


def _get_cryptos() -> List[Plugin]:
    """Get cryptos plugins."""
    return _get_plugins(CRYPTO_PLUGIN_GROUP)


def _get_ledger_apis() -> List[Plugin]:
    """Get ledgers plugins."""
    return _get_plugins(LEDGER_APIS_PLUGIN_GROUP)


def _get_faucet_apis() -> List[Plugin]:
    """Get faucets plugins."""
    return _get_plugins(FAUCET_APIS_PLUGIN_GROUP)


def _iter_plugins() -> Iterator[Plugin]:
    """Iterate over all the plugins."""
    for plugin in itertools.chain(
        _get_cryptos(), _get_ledger_apis(), _get_faucet_apis()
    ):
        yield plugin


def _register_plugin(plugin: Plugin) -> None:
    """Register a plugin to the right registry."""
    register_function = _from_group_to_register_callable[plugin.group]
    register_function(plugin.name, entry_point=plugin.entry_point_path)


def load_all_plugins() -> None:
    """Load all plugins."""
    for plugin in _iter_plugins():
        _register_plugin(plugin)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains tests for 'issue-certificates' command."""
import json
import os
import shutil
from pathlib import Path
from typing import List

import pytest
from aea_ledger_ethereum import EthereumCrypto
from aea_ledger_fetchai import FetchAICrypto

from aea.cli.utils.config import dump_item_config
from aea.helpers.base import CertRequest
from aea.test_tools.test_cases import AEATestCaseEmpty

from tests.conftest import (
    CUR_PATH,
    ETHEREUM_PRIVATE_KEY_FILE,
    ETHEREUM_PRIVATE_KEY_PATH,
    FETCHAI_PRIVATE_KEY_FILE,
    FETCHAI_PRIVATE_KEY_PATH,
)
from tests.data.dummy_connection.connection import DummyConnection


class BaseTestIssueCertificates(AEATestCaseEmpty):
    """Base test class for 'aea issue-certificates' tests."""

    @classmethod
    def setup_class(cls):
        """Set up the class."""
        super().setup_class()

        # add dummy connection
        shutil.copytree(
            os.path.join(CUR_PATH, "data", "dummy_connection"),
            os.path.join(cls.current_agent_context, "connections", "dummy"),
        )
        agent_config = cls.load_agent_config(cls.agent_name)
        agent_config.author = FetchAICrypto.identifier
        agent_config.connections.add(DummyConnection.connection_id)
        dump_item_config(agent_config, Path(cls.current_agent_context))

    @classmethod
    def add_cert_requests(cls, cert_requests: List[CertRequest], connection_name: str):
        """Add certificate requests to a target connection."""
        cls.nested_set_config(
            f"connections.{connection_name}.cert_requests", cert_requests
        )


class TestIssueCertificatesPositive(BaseTestIssueCertificates):
    """Test 'issue-certificates', positive case."""

    @classmethod
    def setup_class(cls):
        """Set up the class."""
        super().setup_class()

        cls.expected_path_1 = os.path.abspath("path_1")
        cls.expected_path_2 = os.path.abspath("path_2")
        cls.cert_id_1 = "cert_id_1"
        cls.cert_id_2 = "cert_id_2"
        cls.cert_request_1 = CertRequest(
            identifier=cls.cert_id_1,
            ledger_id=FetchAICrypto.identifier,
            not_after="2020-01-02",
            not_before="2020-01-01",
            public_key=FetchAICrypto.identifier,
            save_path=cls.expected_path_1,
        )
        cls.cert_request_2 = CertRequest(
            identifier=cls.cert_id_2,
            ledger_id=FetchAICrypto.identifier,
            not_after="2020-01-02",
            not_before="2020-01-01",
            public_key="0xABCDEF123456",
            save_path=cls.expected_path_2,
        )
        cls.add_cert_requests(
            [cls.cert_request_1, cls.cert_request_2], DummyConnection.connection_id.name
        )

        # add fetchai key and connection key
        shutil.copy(
            FETCHAI_PRIVATE_KEY_PATH,
            os.path.join(cls.current_agent_context, FETCHAI_PRIVATE_KEY_FILE),
        )
        cls.add_private_key()
        cls.add_private_key(connection=True)

    def test_issue_certificate(self):
        """Test 'aea issue-certificates' in case of success."""
        result = self.run_cli_command("issue-certificates", cwd=self._get_cwd())
        self._check_signature(self.cert_id_1, self.expected_path_1, result.stdout)
        self._check_signature(self.cert_id_2, self.expected_path_2, result.stdout)

    def _check_signature(self, cert_id, filename, stdout):
        """Check signature has been generated correctly."""
        path = Path(self.current_agent_context, filename)
        assert path.exists()
        signature = path.read_text()

        def is_ascii(s):
            """Check isascii method for all Python 3 versions"""
            return all(ord(c) < 128 for c in s)

        assert is_ascii(signature)
        int(signature, 16)  # this will fail if not hexadecimal

        cert_msg_1 = (
            f"Issuing certificate '{cert_id}' for connection fetchai/dummy:0.1.0..."
        )
        cert_msg_3 = f"Generated signature: '{signature}'"
        cert_msg_2 = f"Dumped certificate '{cert_id}' in '{filename}' for connection fetchai/dummy:0.1.0."
        assert cert_msg_1 in stdout
        assert cert_msg_2 in stdout
        assert cert_msg_3 in stdout


class TestIssueCertificatesWithOverride(TestIssueCertificatesPositive):
    """Test 'issue-certificates' with override configurations."""

    @classmethod
    def setup_class(cls):
        """Set up the class."""
        super().setup_class()

        cls.cert_id_3 = "cert_id_3"
        cls.cert_id_4 = "cert_id_4"
        cls.expected_path_3 = os.path.abspath("path_3")
        cls.expected_path_4 = os.path.abspath("path_4")
        cls.cert_request_3 = CertRequest(
            identifier=cls.cert_id_3,
            ledger_id=EthereumCrypto.identifier,
            not_after="2020-01-02",
            not_before="2020-01-01",
            public_key=EthereumCrypto.identifier,
            save_path=cls.expected_path_3,
        )
        cls.cert_request_4 = CertRequest(
            identifier=cls.cert_id_4,
            ledger_id=EthereumCrypto.identifier,
            not_after="2020-01-02",
            not_before="2020-01-01",
            public_key="0xABCDEF123456",
            save_path=cls.expected_path_4,
        )

        # Add override configurations
        dotted_path = f"connections.{DummyConnection.connection_id.name}.cert_requests"
        json_3 = json.dumps(cls.cert_request_3.json).replace("'", '"')
        json_4 = json.dumps(cls.cert_request_4.json).replace("'", '"')
        new_cert_requests = f"[{json_3}, {json_4}]"
        cls.set_config(dotted_path, new_cert_requests, type_="list")

        # add ethereum key and connection key
        shutil.copy(
            ETHEREUM_PRIVATE_KEY_PATH,
            os.path.join(cls.current_agent_context, ETHEREUM_PRIVATE_KEY_FILE),
        )
        cls.add_private_key(
            ledger_api_id=EthereumCrypto.identifier,
            private_key_filepath=ETHEREUM_PRIVATE_KEY_FILE,
        )
        cls.add_private_key(
            ledger_api_id=EthereumCrypto.identifier,
            private_key_filepath=ETHEREUM_PRIVATE_KEY_FILE,
            connection=True,
        )

    def test_issue_certificate(self):
        """Test 'aea issue-certificates' in case of success."""
        result = self.run_cli_command("issue-certificates", cwd=self._get_cwd())
        self._check_signature(self.cert_id_3, self.expected_path_3, result.stdout)
        self._check_signature(self.cert_id_4, self.expected_path_4, result.stdout)


class TestIssueCertificatesWrongConnectionKey(BaseTestIssueCertificates):
    """Test 'aea issue-certificates' when a bad connection key id is provided."""

    @classmethod
    def setup_class(cls):
        """Set up class."""
        super().setup_class()
        cls.cert_id_1 = "cert_id_1"
        cls.cert_request_1 = CertRequest(
            identifier=cls.cert_id_1,
            ledger_id=FetchAICrypto.identifier,
            not_after="2020-01-02",
            not_before="2020-01-01",
            public_key="bad_ledger_id",
            save_path="path",
        )
        cls.add_cert_requests([cls.cert_request_1], DummyConnection.connection_id.name)

    def test_run(self):
        """Run the test."""
        with pytest.raises(
            Exception,
            match="Cannot find connection private key with id 'bad_ledger_id'",
        ):
            self.run_cli_command("issue-certificates", cwd=self._get_cwd())


class TestIssueCertificatesWrongCryptoKey(BaseTestIssueCertificates):
    """Test 'aea issue-certificates' when a bad crypto key id is provided."""

    @classmethod
    def setup_class(cls):
        """Set up class."""
        super().setup_class()
        cls.cert_id_1 = "cert_id_1"
        cls.cert_request_1 = CertRequest(
            identifier=cls.cert_id_1,
            ledger_id="bad_ledger_id",
            not_after="2020-01-02",
            not_before="2020-01-01",
            public_key=FetchAICrypto.identifier,
            save_path="path",
        )
        cls.add_cert_requests([cls.cert_request_1], DummyConnection.connection_id.name)
        # add fetchai key and connection key
        cls.generate_private_key()
        cls.add_private_key()
        cls.add_private_key(connection=True)

    def test_run(self):
        """Run the test."""
        with pytest.raises(
            Exception, match="Cannot find private key with id 'bad_ledger_id'",
        ):
            self.run_cli_command("issue-certificates", cwd=self._get_cwd())



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the `aea remove-key` sub-command."""
import pytest
from click.exceptions import ClickException

from aea.test_tools.test_cases import AEATestCaseEmpty

from tests.conftest import FETCHAI_PRIVATE_KEY_PATH


class BaseTestRemovePrivateKey(AEATestCaseEmpty):
    """Base test class on removing private keys."""

    WITH_CONNECTION: bool

    @classmethod
    def setup_class(cls):
        """Set up class."""
        super().setup_class()
        cls.add_private_key(
            private_key_filepath=FETCHAI_PRIVATE_KEY_PATH,
            connection=cls.WITH_CONNECTION,
        )

    def test_remove(self):
        """Test remove."""
        result = self.remove_private_key(connection=self.WITH_CONNECTION)
        assert result.exit_code == 0


class TestRemoveCryptoPrivateKey(BaseTestRemovePrivateKey):
    """Test removing a crypto private key."""

    WITH_CONNECTION = False


class TestRemoveConnectionPrivateKey(BaseTestRemovePrivateKey):
    """Test removing a connection private key."""

    WITH_CONNECTION = True


class BaseTestRemovePrivateKeyNegative(AEATestCaseEmpty):
    """Base test class on removing private keys, when key is not present"""

    WITH_CONNECTION: bool
    EXPECTED_ERROR_MSG: str

    def test_remove(self):
        """Test remove."""
        with pytest.raises(ClickException, match=self.EXPECTED_ERROR_MSG):
            self.remove_private_key(connection=self.WITH_CONNECTION)


class TestRemoveCryptoPrivateKeyNegative(BaseTestRemovePrivateKeyNegative):
    """Test removing a crypto private key."""

    WITH_CONNECTION = False
    EXPECTED_ERROR_MSG = "There is no key registered with id fetchai."


class TestRemoveConnectionPrivateKeyNegative(BaseTestRemovePrivateKeyNegative):
    """Test removing a connection private key."""

    WITH_CONNECTION = True
    EXPECTED_ERROR_MSG = "There is no connection key registered with id fetchai."



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
Generate all the protocols from their specifications.

This script takes all the protocol specification (scraped from the protocol README)
and calls the `aea generate protocol` command.

Currently, it does a lot of assumptions, and might not be useful for
all use cases. However, with not much work, can be customized to achieve
the desired outcomes.

It requires the `aea` package, `black` and `isort` tools.
"""
import argparse
import logging
import os
import pprint
import re
import shutil
import subprocess  # nosec
import sys
import tempfile
from io import StringIO
from itertools import chain
from operator import methodcaller
from pathlib import Path
from typing import Any, Iterator, List, Match, Optional, Tuple, cast

import semver

from aea.cli.registry.utils import download_file, extract, request_api
from aea.common import JSONLike
from aea.configurations.base import ComponentType, ProtocolConfig, ProtocolSpecification
from aea.configurations.constants import DEFAULT_PROTOCOL_CONFIG_FILE, LIBPROTOC_VERSION
from aea.configurations.data_types import PackageId, PublicId
from aea.configurations.loader import (
    ConfigLoader,
    ConfigLoaders,
    load_component_configuration,
)
from scripts.common import check_working_tree_is_dirty


SPECIFICATION_REGEX = re.compile(r"(---\nname.*\.\.\.)", re.DOTALL)
CUSTOM_TYPE_MODULE_NAME = "custom_types.py"
README_FILENAME = "README.md"
PACKAGES_DIR = Path("packages")
TEST_DATA = Path("tests", "data").absolute()
PROTOCOLS_PLURALS = "protocols"
ROOT_DIR = Path(".").absolute()


def subdirs(path: Path) -> Iterator[Path]:
    """Get subdirectories of a path."""
    return filter(methodcaller("is_dir"), path.iterdir())


def find_protocols_in_local_registry() -> Iterator[Path]:
    """Find all protocols in local registry."""
    authors = subdirs(PACKAGES_DIR)
    component_parents = chain(*map(subdirs, authors))
    protocols_parent = filter(lambda p: p.name == PROTOCOLS_PLURALS, component_parents)
    protocols = chain(*map(subdirs, protocols_parent))
    return map(methodcaller("absolute"), protocols)


def _setup_logger() -> logging.Logger:
    """Set up the logger."""
    FORMAT = "[%(asctime)s][%(levelname)s] %(message)s"
    logging.basicConfig(format=FORMAT)
    logger_ = logging.getLogger("generate_all_protocols")
    logger_.setLevel(logging.INFO)
    return logger_


logger = _setup_logger()


def log(message: str, level: int = logging.INFO) -> None:
    """Produce a logging message."""
    logger.log(level, message)


def enforce(condition: bool, message: str = "") -> None:
    """Custom assertion."""
    if not condition:
        raise AssertionError(message)


def run_cli(*args: Any, **kwargs: Any) -> None:
    """Run a CLI command."""
    log(f"Calling command {args} with kwargs {kwargs}")
    return_code = subprocess.check_call(args, **kwargs)  # nosec
    enforce(
        return_code == 0,
        f"Return code of {pprint.pformat(args)} is {return_code} != 0.",
    )


def run_aea(*args: Any, **kwargs: Any) -> None:
    """
    Run an AEA command.

    :param args: the AEA command
    :param kwargs: keyword arguments to subprocess function
    :return: None
    """
    run_cli(sys.executable, "-m", "aea.cli", *args, **kwargs)


class AEAProject:
    """A context manager class to create and delete an AEA project."""

    old_cwd: str
    temp_dir: str

    def __init__(self, name: str = "my_aea", parent_dir: Optional[str] = None):
        """
        Initialize an AEA project.

        :param name: the name of the AEA project.
        """
        self.name = name
        self.parent_dir = parent_dir

    def __enter__(self) -> None:
        """Create and enter into the project."""
        self.old_cwd = os.getcwd()
        self.temp_dir = tempfile.mkdtemp(dir=self.parent_dir)
        os.chdir(self.temp_dir)

        run_aea("create", "--local", "--empty", self.name, "--author", "fetchai")
        os.chdir(self.name)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        """Exit the context manager."""
        os.chdir(self.old_cwd)
        shutil.rmtree(self.temp_dir)


def _load_protocol_specification_from_string(
    specification_content: str,
) -> ProtocolSpecification:
    """Load a protocol specification from string."""
    file = StringIO(initial_value=specification_content)
    config_loader = ConfigLoader(
        "protocol-specification_schema.json", ProtocolSpecification
    )
    protocol_spec = config_loader.load_protocol_specification(file)
    return protocol_spec


def _get_protocol_specification_from_readme(package_path: Path) -> str:
    """Get the protocol specification from the package README."""
    log(f"Get protocol specification from README {package_path}")
    readme = package_path / "README.md"
    readme_content = readme.read_text()
    enforce(
        "## Specification" in readme_content,
        f"Cannot find specification section in {package_path}",
    )

    search_result = SPECIFICATION_REGEX.search(readme_content)
    enforce(
        search_result is not None,
        f"Cannot find specification section in README of {package_path}",
    )
    specification_content = cast(Match, search_result).group(0)
    # just for validation of the parsed string
    _load_protocol_specification_from_string(specification_content)
    return specification_content


def _save_specification_in_temporary_file(
    name: str, specification_content: str
) -> None:
    """
    Save the specification in a temporary file.

    :param name: the name of the package.
    :param specification_content: the specification content.
    :return: None
    """
    # here, the cwd is the temporary AEA project
    # hence, we are writing in a temporary directory
    spec_path = Path("..", name + ".yaml")
    log(f"Save specification '{name}' in temporary file {spec_path}")
    spec_path.write_text(specification_content)


def _generate_protocol(package_path: Path) -> None:
    """
    Generate the protocol.

    :param package_path: package to the path.
    :return: None
    """
    cmd = ["generate", "protocol", os.path.join("..", package_path.name) + ".yaml"]
    log(f"Generate the protocol. Command: {pprint.pformat(cmd)}")
    run_aea(*cmd)


def run_isort_and_black(directory: Path, **kwargs: Any) -> None:
    """Run black and isort against a directory."""
    run_cli(
        sys.executable, "-m", "black", "--verbose", str(directory.absolute()), **kwargs,
    )
    run_cli(
        sys.executable,
        "-m",
        "isort",
        "--settings-path",
        "setup.cfg",
        "--verbose",
        str(directory.absolute()),
        **kwargs,
    )


def replace_in_directory(name: str, replacement_pairs: List[Tuple[str, str]]) -> None:
    """
    Replace text in directory.

    :param name: the protocol name.
    :param replacement_pairs: a list of pairs of strings (to_replace, replacement).
    :return: None
    """
    log(f"Replace prefix of import statements in directory '{name}'")
    package_dir = Path(PROTOCOLS_PLURALS, name)
    for submodule in package_dir.rglob("*.py"):
        log(f"Process submodule {submodule.relative_to(package_dir)}")
        for to_replace, replacement in replacement_pairs:
            if to_replace not in submodule.read_text():
                continue
            submodule.write_text(submodule.read_text().replace(to_replace, replacement))


def _fix_generated_protocol(package_path: Path) -> None:
    """
    Fix the generated protocol.

    That means:
    - replacing the prefix of import statements for default protocols;
    - restore the original custom types, if any.
    - copy the README, if any.

    :param package_path: path to the protocol package.
                         Used also to recover the protocol name.
    :return: None
    """
    log(f"Restore original custom types in {package_path}")
    custom_types_module = package_path / CUSTOM_TYPE_MODULE_NAME
    if custom_types_module.exists():
        file_to_replace = Path(
            PROTOCOLS_PLURALS, package_path.name, CUSTOM_TYPE_MODULE_NAME
        )
        file_to_replace.write_text(custom_types_module.read_text())

    package_readme_file = package_path / README_FILENAME
    if package_readme_file.exists():
        log(f"Copy the README {package_readme_file} into the new generated protocol.")
        shutil.copyfile(
            package_readme_file,
            Path(PROTOCOLS_PLURALS, package_path.name, README_FILENAME),
        )


def _update_original_protocol(package_path: Path) -> None:
    """
    Update the original protocol.

    :param package_path: the path to the original package.
                         Used to recover the protocol name.
    :return: None
    """
    log(f"Copy the new protocol into the original directory {package_path}")
    shutil.rmtree(package_path)
    shutil.copytree(Path(PROTOCOLS_PLURALS, package_path.name), package_path)


def _fingerprint_protocol(name: str) -> None:
    """Fingerprint the generated (and modified) protocol."""
    log(f"Fingerprint the generated (and modified) protocol '{name}'")
    protocol_config = load_component_configuration(
        ComponentType.PROTOCOL,
        Path(PROTOCOLS_PLURALS, name),
        skip_consistency_check=True,
    )
    run_aea("fingerprint", "protocol", str(protocol_config.public_id))


def _process_packages_protocol(package_path: Path) -> None:
    """
    Process protocol package from local registry.

    It means:
    - extract protocol specification from README
    - generate the protocol in the current AEA project
    - fix the generated protocol (e.g. import prefixed, custom types, ...)
    - update the original protocol with the newly generated one.

    It assumes the working directory is an AEA project.

    :param package_path: path to the package.
    :return: None
    """
    specification_content = _get_protocol_specification_from_readme(package_path)
    _save_specification_in_temporary_file(package_path.name, specification_content)
    _generate_protocol(package_path)
    _fix_generated_protocol(package_path)
    run_isort_and_black(Path(PROTOCOLS_PLURALS, package_path.name), cwd=str(ROOT_DIR))
    _fingerprint_protocol(package_path.name)
    _update_original_protocol(package_path)


def _check_preliminaries() -> None:
    """Check that the required packages are installed."""
    try:
        import aea  # noqa: F401  # pylint: disable=import-outside-toplevel,unused-import
    except ModuleNotFoundError:
        enforce(False, "'aea' package not installed.")
    enforce(shutil.which("black") is not None, "black command line tool not found.")
    enforce(shutil.which("isort") is not None, "isort command line tool not found.")
    enforce(shutil.which("protoc") is not None, "protoc command line tool not found.")
    result = subprocess.run(  # nosec
        ["protoc", "--version"], stdout=subprocess.PIPE, check=True
    )
    result_str = result.stdout.decode("utf-8")
    enforce(
        LIBPROTOC_VERSION in result_str,
        f"Invalid version for protoc. Found: {result_str}. Required: {LIBPROTOC_VERSION}.",
    )


def _process_test_protocol(specification: Path, package_path: Path) -> None:
    """
    Process a test protocol.

    :param specification: path to specification.
    :param package_path: the output directory.
    :return: None
    """
    specification_content = specification.read_text()
    _save_specification_in_temporary_file(package_path.name, specification_content)
    _generate_protocol(package_path)
    _fix_generated_protocol(package_path)
    replacements = [
        (
            f"from packages.fetchai.protocols.{package_path.name}",
            f"from tests.data.generator.{package_path.name}",
        )
    ]
    replace_in_directory(package_path.name, replacements)
    run_isort_and_black(Path(PROTOCOLS_PLURALS, package_path.name), cwd=str(ROOT_DIR))
    _fingerprint_protocol(package_path.name)
    _update_original_protocol(package_path)


def download_package(package_id: PackageId, destination_path: str) -> None:
    """Download a package into a directory."""
    api_path = f"/{package_id.package_type.to_plural()}/{package_id.author}/{package_id.name}/{package_id.public_id.LATEST_VERSION}"
    resp = cast(JSONLike, request_api("GET", api_path))
    file_url = cast(str, resp["file"])
    filepath = download_file(file_url, destination_path)
    extract(filepath, destination_path)


def _bump_protocol_specification_id(
    package_path: Path, configuration: ProtocolConfig
) -> None:
    """Bump spec id version."""
    spec_id: PublicId = configuration.protocol_specification_id  # type: ignore
    old_version = semver.VersionInfo.parse(spec_id.version)
    new_version = str(old_version.bump_minor())
    new_spec_id = PublicId(spec_id.author, spec_id.name, new_version)
    configuration.protocol_specification_id = new_spec_id
    with (package_path / DEFAULT_PROTOCOL_CONFIG_FILE).open("w") as file_output:
        ConfigLoaders.from_package_type(configuration.package_type).dump(
            configuration, file_output
        )


def _bump_protocol_specification_id_if_needed(package_path: Path) -> None:
    """
    Check if protocol specification id needs to be bumped.

    Workflow:
    - extract protocol specification file from README
    - download latest protocol id and extract its protocol specification as above
    - if different, bump protocol specification version, else don't.

    :param package_path: path to the protocol package.
    :return: None
    """
    # extract protocol specification file from README
    current_specification_content = _get_protocol_specification_from_readme(
        package_path
    )

    # download latest protocol id and extract its protocol specification as above
    configuration: ProtocolConfig = cast(
        ProtocolConfig,
        load_component_configuration(ComponentType.PROTOCOL, package_path),
    )
    temp_directory = Path(tempfile.mkdtemp())
    download_package(configuration.package_id, str(temp_directory))
    downloaded_package_directory = temp_directory / configuration.name
    old_specification_content = _get_protocol_specification_from_readme(
        downloaded_package_directory
    )
    old_configuration: ProtocolConfig = cast(
        ProtocolConfig,
        load_component_configuration(
            ComponentType.PROTOCOL,
            downloaded_package_directory,
            skip_consistency_check=True,
        ),
    )

    # if different, bump protocol specification version, else don't.
    public_id_version_is_newer = (
        old_configuration.public_id.package_version  # type: ignore
        <= configuration.public_id.package_version
    )
    content_is_different = current_specification_content != old_specification_content
    if public_id_version_is_newer and content_is_different:
        log(
            f"Bumping protocol specification id from '{old_configuration.protocol_specification_id}' to '{configuration.protocol_specification_id}'"
        )
        _bump_protocol_specification_id(package_path, configuration)
        return
    log(
        "Protocol specification id not bumped - content is not different, or version is not newer."
    )


def main() -> None:
    """Run the script."""
    _check_preliminaries()

    all_protocols = list(find_protocols_in_local_registry())

    with AEAProject():
        log("=" * 100)
        _process_test_protocol(
            TEST_DATA / "sample_specification.yaml",
            TEST_DATA / "generator" / "t_protocol",
        )
        log("=" * 100)
        _process_test_protocol(
            TEST_DATA / "sample_specification_no_custom_types.yaml",
            TEST_DATA / "generator" / "t_protocol_no_ct",
        )
        for package_path in all_protocols:
            log("=" * 100)
            log(f"Processing protocol at path {package_path}")
            _bump_protocol_specification_id_if_needed(package_path)
            _process_packages_protocol(package_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser("generate_all_protocols")
    parser.add_argument(
        "--check-clean", action="store_true", help="Check if the working tree is clean."
    )
    arguments = parser.parse_args()

    main()

    if arguments.check_clean:
        check_working_tree_is_dirty()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: yoti.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="yoti.proto",
    package="aea.fetchai.yoti.v0_1_0",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=b'\n\nyoti.proto\x12\x17\x61\x65\x61.fetchai.yoti.v0_1_0"\xaf\x04\n\x0bYotiMessage\x12H\n\x05\x65rror\x18\x05 \x01(\x0b\x32\x37.aea.fetchai.yoti.v0_1_0.YotiMessage.Error_PerformativeH\x00\x12T\n\x0bget_profile\x18\x06 \x01(\x0b\x32=.aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_PerformativeH\x00\x12L\n\x07profile\x18\x07 \x01(\x0b\x32\x39.aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_PerformativeH\x00\x1aL\n\x18Get_Profile_Performative\x12\r\n\x05token\x18\x01 \x01(\t\x12\x13\n\x0b\x64otted_path\x18\x02 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x03 \x03(\t\x1a\x96\x01\n\x14Profile_Performative\x12Q\n\x04info\x18\x01 \x03(\x0b\x32\x43.aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a;\n\x12\x45rror_Performative\x12\x12\n\nerror_code\x18\x01 \x01(\x05\x12\x11\n\terror_msg\x18\x02 \x01(\tB\x0e\n\x0cperformativeb\x06proto3',
)


_YOTIMESSAGE_GET_PROFILE_PERFORMATIVE = _descriptor.Descriptor(
    name="Get_Profile_Performative",
    full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="token",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_Performative.token",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dotted_path",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_Performative.dotted_path",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="args",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_Performative.args",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=293,
    serialized_end=369,
)

_YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY = _descriptor.Descriptor(
    name="InfoEntry",
    full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.InfoEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.InfoEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.InfoEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=479,
    serialized_end=522,
)

_YOTIMESSAGE_PROFILE_PERFORMATIVE = _descriptor.Descriptor(
    name="Profile_Performative",
    full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="info",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.info",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[_YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY,],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=372,
    serialized_end=522,
)

_YOTIMESSAGE_ERROR_PERFORMATIVE = _descriptor.Descriptor(
    name="Error_Performative",
    full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Error_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="error_code",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Error_Performative.error_code",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="error_msg",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Error_Performative.error_msg",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=524,
    serialized_end=583,
)

_YOTIMESSAGE = _descriptor.Descriptor(
    name="YotiMessage",
    full_name="aea.fetchai.yoti.v0_1_0.YotiMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="error",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.error",
            index=0,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="get_profile",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.get_profile",
            index=1,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="profile",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.profile",
            index=2,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[
        _YOTIMESSAGE_GET_PROFILE_PERFORMATIVE,
        _YOTIMESSAGE_PROFILE_PERFORMATIVE,
        _YOTIMESSAGE_ERROR_PERFORMATIVE,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="performative",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.performative",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=40,
    serialized_end=599,
)

_YOTIMESSAGE_GET_PROFILE_PERFORMATIVE.containing_type = _YOTIMESSAGE
_YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY.containing_type = (
    _YOTIMESSAGE_PROFILE_PERFORMATIVE
)
_YOTIMESSAGE_PROFILE_PERFORMATIVE.fields_by_name[
    "info"
].message_type = _YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY
_YOTIMESSAGE_PROFILE_PERFORMATIVE.containing_type = _YOTIMESSAGE
_YOTIMESSAGE_ERROR_PERFORMATIVE.containing_type = _YOTIMESSAGE
_YOTIMESSAGE.fields_by_name["error"].message_type = _YOTIMESSAGE_ERROR_PERFORMATIVE
_YOTIMESSAGE.fields_by_name[
    "get_profile"
].message_type = _YOTIMESSAGE_GET_PROFILE_PERFORMATIVE
_YOTIMESSAGE.fields_by_name["profile"].message_type = _YOTIMESSAGE_PROFILE_PERFORMATIVE
_YOTIMESSAGE.oneofs_by_name["performative"].fields.append(
    _YOTIMESSAGE.fields_by_name["error"]
)
_YOTIMESSAGE.fields_by_name["error"].containing_oneof = _YOTIMESSAGE.oneofs_by_name[
    "performative"
]
_YOTIMESSAGE.oneofs_by_name["performative"].fields.append(
    _YOTIMESSAGE.fields_by_name["get_profile"]
)
_YOTIMESSAGE.fields_by_name[
    "get_profile"
].containing_oneof = _YOTIMESSAGE.oneofs_by_name["performative"]
_YOTIMESSAGE.oneofs_by_name["performative"].fields.append(
    _YOTIMESSAGE.fields_by_name["profile"]
)
_YOTIMESSAGE.fields_by_name["profile"].containing_oneof = _YOTIMESSAGE.oneofs_by_name[
    "performative"
]
DESCRIPTOR.message_types_by_name["YotiMessage"] = _YOTIMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

YotiMessage = _reflection.GeneratedProtocolMessageType(
    "YotiMessage",
    (_message.Message,),
    {
        "Get_Profile_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Profile_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _YOTIMESSAGE_GET_PROFILE_PERFORMATIVE,
                "__module__": "yoti_pb2"
                # @@protoc_insertion_point(class_scope:aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_Performative)
            },
        ),
        "Profile_Performative": _reflection.GeneratedProtocolMessageType(
            "Profile_Performative",
            (_message.Message,),
            {
                "InfoEntry": _reflection.GeneratedProtocolMessageType(
                    "InfoEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY,
                        "__module__": "yoti_pb2"
                        # @@protoc_insertion_point(class_scope:aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.InfoEntry)
                    },
                ),
                "DESCRIPTOR": _YOTIMESSAGE_PROFILE_PERFORMATIVE,
                "__module__": "yoti_pb2"
                # @@protoc_insertion_point(class_scope:aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _YOTIMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "yoti_pb2"
                # @@protoc_insertion_point(class_scope:aea.fetchai.yoti.v0_1_0.YotiMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _YOTIMESSAGE,
        "__module__": "yoti_pb2"
        # @@protoc_insertion_point(class_scope:aea.fetchai.yoti.v0_1_0.YotiMessage)
    },
)
_sym_db.RegisterMessage(YotiMessage)
_sym_db.RegisterMessage(YotiMessage.Get_Profile_Performative)
_sym_db.RegisterMessage(YotiMessage.Profile_Performative)
_sym_db.RegisterMessage(YotiMessage.Profile_Performative.InfoEntry)
_sym_db.RegisterMessage(YotiMessage.Error_Performative)


_YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the strategy class of the confirmation aw2 skill."""

import datetime
import logging
from pathlib import Path
from typing import cast
from unittest.mock import Mock, patch

import pytest

from packages.fetchai.skills.confirmation_aw2.registration_db import RegistrationDB
from packages.fetchai.skills.confirmation_aw2.strategy import Strategy

from tests.conftest import ROOT_DIR
from tests.test_packages.test_skills.test_confirmation_aw2.intermediate_class import (
    ConfirmationAW2TestCase,
)


class TestStrategy(ConfirmationAW2TestCase):
    """Test Strategy of confirmation aw2."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "confirmation_aw2")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()

        cls.minimum_hours_between_txs = 4
        cls.minimum_minutes_since_last_attempt = 2
        cls.strategy = Strategy(
            aw1_aea="some_aw1_aea",
            mininum_hours_between_txs=cls.minimum_hours_between_txs,
            minimum_minutes_since_last_attempt=cls.minimum_minutes_since_last_attempt,
            name="strategy",
            skill_context=cls._skill.skill_context,
        )

        cls.address = "some_address"
        cls.info = {
            "ethereum_address": "some_value",
            "signature_of_ethereum_address": "some_signature_of_ethereum_address",
            "signature_of_fetchai_address": "some_signature_of_fetchai_address",
            "developer_handle": "some_developer_handle",
            "tweet": "some_tweet",
        }
        cls.logger = cls._skill.skill_context.logger
        cls.db = cast(RegistrationDB, cls._skill.skill_context.registration_db)

        cls.counterparty = "couterparty_1"

    def test__init__i(self):
        """Test the __init__ of Strategy class."""
        assert self.strategy.aw1_aea == self.aw1_aea
        assert self.strategy.minimum_hours_between_txs == self.minimum_hours_between_txs
        assert (
            self.strategy.minimum_minutes_since_last_attempt
            == self.minimum_minutes_since_last_attempt
        )

    def test__init__ii(self):
        """Test the __init__ of Strategy class where aw1_aea is None."""
        with pytest.raises(ValueError, match="aw1_aea must be provided!"):
            Strategy(
                aw1_aea=None,
                mininum_hours_between_txs=self.minimum_hours_between_txs,
                minimum_minutes_since_last_attempt=self.minimum_minutes_since_last_attempt,
                name="strategy",
                skill_context=self.skill.skill_context,
            )

    def test_get_acceptable_counterparties(self):
        """Test the get_acceptable_counterparties method of the Strategy class."""
        # setup
        couterparties = ("couterparty_1", "couterparty_2", "couterparty_3")
        is_valid_counterparty = [True, False, True]

        # operation
        with patch.object(
            self.strategy, "is_valid_counterparty", side_effect=is_valid_counterparty
        ):
            actual_acceptable_counterparties = self.strategy.get_acceptable_counterparties(
                couterparties
            )

        # after
        assert actual_acceptable_counterparties == ("couterparty_1", "couterparty_3")

    def test_is_enough_time_since_last_attempt_i(self):
        """Test the is_enough_time_since_last_attempt method of the Strategy class where now IS greater than last attempt + min minutes."""
        # setup
        counterparty_last_attempt_time_str = "2020-12-22 20:30:00.000000"
        counterparty_last_attempt_time = datetime.datetime.strptime(
            counterparty_last_attempt_time_str, "%Y-%m-%d %H:%M:%S.%f"
        )

        mocked_now_greater_than_last_plus_minimum = "2020-12-22 20:33:00.000000"
        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = datetime.datetime.strptime(
            mocked_now_greater_than_last_plus_minimum, "%Y-%m-%d %H:%M:%S.%f"
        )
        self.strategy.last_attempt = {self.counterparty: counterparty_last_attempt_time}

        # operation
        with patch("datetime.datetime", new=datetime_mock):
            is_enough_time = self.strategy.is_enough_time_since_last_attempt(
                self.counterparty
            )

        # after
        assert is_enough_time is True

    def test_is_enough_time_since_last_attempt_ii(self):
        """Test the is_enough_time_since_last_attempt method of the Strategy class where now is NOT greater than last attempt + min minutes."""
        # setup
        counterparty_last_attempt_time_str = "2020-12-22 20:30:00.000000"
        counterparty_last_attempt_time = datetime.datetime.strptime(
            counterparty_last_attempt_time_str, "%Y-%m-%d %H:%M:%S.%f"
        )

        mocked_now_less_than_last_plus_minimum = "2020-12-22 20:31:00.000000"
        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = datetime.datetime.strptime(
            mocked_now_less_than_last_plus_minimum, "%Y-%m-%d %H:%M:%S.%f"
        )
        self.strategy.last_attempt = {self.counterparty: counterparty_last_attempt_time}

        # operation
        with patch("datetime.datetime", new=datetime_mock):
            is_enough_time = self.strategy.is_enough_time_since_last_attempt(
                self.counterparty
            )

        # after
        assert is_enough_time is False

    def test_is_enough_time_since_last_attempt_iii(self):
        """Test the is_enough_time_since_last_attempt method of the Strategy class where now counterparty is NOT in last_attempt."""
        # setup
        self.strategy.last_attempt = {}

        # operation
        is_enough_time = self.strategy.is_enough_time_since_last_attempt(
            self.counterparty
        )

        # after
        assert is_enough_time is True

    def test_is_valid_counterparty_i(self):
        """Test the is_valid_counterparty method of the Strategy class where is_registered is False."""
        # operation
        with patch.object(self.db, "is_registered", return_value=False):
            with patch.object(self.logger, "log") as mock_logger:
                is_valid = self.strategy.is_valid_counterparty(self.counterparty)

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"Invalid counterparty={self.counterparty}, not registered!",
        )
        assert is_valid is False

    def test_is_valid_counterparty_ii(self):
        """Test the is_valid_counterparty method of the Strategy class where is_enough_time_since_last_attempt is False."""
        # operation
        with patch.object(self.db, "is_registered", return_value=True):
            with patch.object(
                self.strategy, "is_enough_time_since_last_attempt", return_value=False
            ):
                with patch.object(self.logger, "log") as mock_logger:
                    is_valid = self.strategy.is_valid_counterparty(self.counterparty)

        # after
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"Not enough time since last attempt for counterparty={self.counterparty}!",
        )
        assert is_valid is False

    def test_is_valid_counterparty_iii(self):
        """Test the is_valid_counterparty method of the Strategy class where is_allowed_to_trade is False."""
        # operation
        with patch.object(self.db, "is_registered", return_value=True):
            with patch.object(
                self.strategy, "is_enough_time_since_last_attempt", return_value=True
            ):
                with patch.object(self.db, "is_allowed_to_trade", return_value=False):
                    is_valid = self.strategy.is_valid_counterparty(self.counterparty)

        # after
        assert is_valid is False

    def test_is_valid_counterparty_iv(self):
        """Test the is_valid_counterparty method of the Strategy class where it succeeds."""
        # operation
        with patch.object(self.db, "is_registered", return_value=True):
            with patch.object(
                self.strategy, "is_enough_time_since_last_attempt", return_value=True
            ):
                with patch.object(self.db, "is_allowed_to_trade", return_value=True):
                    is_valid = self.strategy.is_valid_counterparty(self.counterparty)

        # after
        assert is_valid is True

    def test_successful_trade_with_counterparty(self):
        """Test the successful_trade_with_counterparty method of the Strategy class."""
        # setup
        data = {"some_key_1": "some_value_1", "some_key_2": "some_value_2"}

        mocked_now_str = "2020-12-22 20:33:00.000000"
        mock_now = datetime.datetime.strptime(mocked_now_str, "%Y-%m-%d %H:%M:%S.%f")
        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = mock_now

        # operation
        with patch.object(self.db, "set_trade") as mock_set_trade:
            with patch("datetime.datetime", new=datetime_mock):
                with patch.object(self.logger, "log") as mock_logger:
                    self.strategy.successful_trade_with_counterparty(
                        self.counterparty, data
                    )

        # after
        mock_set_trade.assert_any_call(self.counterparty, mock_now, data)

        mock_logger.assert_any_call(
            logging.INFO,
            f"Successful trade with={self.counterparty}. Data acquired={data}!",
        )

    def test_register_counterparty(self):
        """Test the register_counterparty method of the Strategy class."""
        # setup
        developer_handle = "some_developer_handle"

        # operation
        with patch.object(self.db, "set_registered") as mock_set_registered:
            self.strategy.register_counterparty(self.counterparty, developer_handle)

        # after
        mock_set_registered.assert_any_call(self.counterparty, developer_handle)
